from . import io
from . import plotting
from . import corrections
from . import utils
from . import tests
from . import discovery
from . import retrievals
from . import qc
from ._version import get_versions

# No more pandas warnings
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

# Version for source builds
vdict = get_versions()
__version__ = vdict["version"]

# Version for releases
# __version__ = "0.1"



"""
===============================
act.retrievals (act.retrievals)
===============================

.. currentmodule:: act.retrievals

This module contains various retrievals for datsets.

.. autosummary::
    :toctree: generated/

    aeri2irt
    calculate_dsh_from_dsdh_sdn
    calculate_irradiance_stats
    calculate_net_radiation
    calculate_longwave_radiation
    calculate_precipitable_water
    calculate_stability_indicies
    compute_winds_from_ppi
    generic_sobel_cbh
    sst_from_irt
    sum_function_irt
"""

from .stability_indices import calculate_stability_indicies
from .cbh import generic_sobel_cbh
from .pwv_calc import calculate_precipitable_water
from .doppler_lidar import compute_winds_from_ppi
from .aeri import aeri2irt
from .irt import sst_from_irt
from .irt import sum_function_irt
from .radiation import calculate_dsh_from_dsdh_sdn
from .radiation import calculate_irradiance_stats
from .radiation import calculate_net_radiation
from .radiation import calculate_longwave_radiation



"""
act.plotting.common
-------------------

Functions common between plotting modules.

"""

import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter


def parse_ax(ax):
    """
    Parses the given matplotlib axis.

    Parameters
    ----------
    ax : matplotlib axis
        The matplotlib axis to be parsed. Set to none to get the current axis.

    Returns
    -------
    ax : matplotlib axis
        The target matplotlib axis.

    """
    if ax is None:
        ax = plt.gca()
    return ax


def parse_ax_fig(ax, fig):
    """
    Parses the given matplotlib axis and figure.

    Parameters
    ----------
    ax : matplotlib axis
        The matplotlib axis to be parsed. Set to None to get the current axis.
    fig : matplotlib fig
        The matplotlib figure to be parsed.
        Set to None to get the current figure.

    Returns
    -------
    ax : matplotlib axis
        The target matplotlib axis.
    fig : matplotlib figure
        The target matplotlib figure.

    """
    if ax is None:
        ax = plt.gca()
    if fig is None:
        fig = plt.gcf()
    return ax, fig


def get_date_format(days, day_to_hour_threshold=3):
    """
    Returns the DateFormatter object to use for the given number of days.

    Parameters
    ----------
    days : float
        The number of days we are plotting.
    day_to_hour_threshold : float
        If the dataset is under this threshold long, format the x axis ticks
        by hour instead of date.

    Returns
    -------
    myFmt : matplotlib DateFormatter
        The DateFormatter object to use for the plot.

    """
    # Set format for time axis if needed
    if days > day_to_hour_threshold:
        myFmt = DateFormatter('%b %d')
    else:
        myFmt = DateFormatter('%H:%M')

    return myFmt



import xarray as xr
import json
import time
import pandas as pd
import numpy as np
import warnings

from six import StringIO
from datetime import datetime

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


def get_asos(time_window, lat_range=None,
             lon_range=None, station=None):
    """
    Returns all of the station observations from the Iowa Mesonet from either
    a given latitude and longitude window or a given station code.

    Parameters
    ----------
    time_window: tuple
        A 2 member list or tuple containing the start and end times. The
        times must be python datetimes.
    lat_range: tuple
        The latitude window to grab all of the ASOS observations from.
    lon_range: tuple
        The longitude window to grab all of the ASOS observations from.
    station: str
        The station ID to grab the ASOS observations from.

    Returns
    -------
    asos_ds: dict of xarray datasets
        A dictionary of ACT datasets whose keys are the ASOS station IDs.

    Examples
    --------
    If you want to obtain timeseries of ASOS observations for Chicago O'Hare
    Airport, simply do::

        $ time_window = [datetime(2020, 2, 4, 2, 0), datetime(2020, 2, 10, 10, 0)]
        $ station = "KORD"
        $ my_asoses = act.discovery.get_asos(time_window, station="ORD")
    """

    # First query the database for all of the JSON info for every station
    # Only add stations whose lat/lon are within the Grid's boundaries
    regions = """AF AL_ AI_ AQ_ AG_ AR_ AK AL AM_
        AO_ AS_ AR AW_ AU_ AT_
        AZ_ BA_ BE_ BB_ BG_ BO_ BR_ BF_
        BT_ BS_ BI_ BM_ BB_ BY_ BZ_ BJ_ BW_ AZ CA CA_AB
        CA_BC CD_ CK_ CF_ CG_ CL_ CM_ CO CO_ CN_ CR_ CT
        CU_ CV_ CY_ CZ_ DE DK_ DJ_ DM_ DO_
        DZ EE_ ET_ FK_ FM_ FJ_ FI_ FR_ GF_ PF_
        GA_ GM_ GE_ DE_ GH_ GI_ KY_ GB_ GR_ GL_ GD_
        GU_ GT_ GN_ GW_ GY_ HT_ HN_ HK_ HU_ IS_ IN_
        ID_ IR_ IQ_ IE_ IL_ IT_ CI_ JM_ JP_
        JO_ KZ_ KE_ KI_ KW_ LA_ LV_ LB_ LS_ LR_ LY_
        LT_ LU_ MK_ MG_ MW_ MY_ MV_ ML_ CA_MB
        MH_ MR_ MU_ YT_ MX_ MD_ MC_ MA_ MZ_ MM_ NA_ NP_
        AN_ NL_ CA_NB NC_ CA_NF NF_ NI_
        NE_ NG_ MP_ KP_ CA_NT NO_ CA_NS CA_NU OM_
        CA_ON PK_ PA_ PG_ PY_ PE_ PH_ PN_ PL_
        PT_ CA_PE PR_ QA_ CA_QC RO_ RU_RW_ SH_ KN_
        LC_ VC_ WS_ ST_ CA_SK SA_ SN_ RS_ SC_
        SL_ SG_ SK_ SI_ SB_ SO_ ZA_ KR_ ES_ LK_ SD_ SR_
        SZ_ SE_ CH_ SY_ TW_ TJ_ TZ_ TH_
        TG_ TO_ TT_ TU TN_ TR_ TM_ UG_ UA_ AE_ UN_ UY_
        UZ_ VU_ VE_ VN_ VI_ YE_ CA_YT ZM_ ZW_
        EC_ EG_ FL GA GQ_ HI HR_ IA ID IL IO_ IN KS
        KH_ KY KM_ LA MA MD ME
        MI MN MO MS MT NC ND NE NH NJ NM NV NY OH OK
        OR PA RI SC SV_ SD TD_ TN TX UT VA VT VG_
        WA WI WV WY"""

    networks = ["AWOS"]
    metadata_list = {}
    if lat_range is not None and lon_range is not None:
        lon_min, lon_max = lon_range
        lat_min, lat_max = lat_range
        for region in regions.split():
            networks.append("%s_ASOS" % (region,))

        site_list = []
        for network in networks:
            # Get metadata
            uri = ("https://mesonet.agron.iastate.edu/"
                   "geojson/network/%s.geojson") % (network,)
            data = urlopen(uri)
            jdict = json.load(data)
            for site in jdict["features"]:
                lat = site["geometry"]["coordinates"][1]
                lon = site["geometry"]["coordinates"][0]
                if lat >= lat_min and lat <= lat_max:
                    if lon >= lon_min and lon <= lon_max:
                        station_metadata_dict = {}
                        station_metadata_dict["site_latitude"] = lat
                        station_metadata_dict["site_longitude"] = lat
                        for my_keys in site["properties"]:
                            station_metadata_dict[my_keys] = site["properties"][my_keys]
                        metadata_list[site["properties"]["sid"]] = station_metadata_dict
                        site_list.append(site["properties"]["sid"])
    elif station is not None:
        site_list = [station]
        for region in regions.split():
            networks.append("%s_ASOS" % (region,))
        for network in networks:
            # Get metadata
            uri = ("https://mesonet.agron.iastate.edu/"
                   "geojson/network/%s.geojson") % (network,)
            data = urlopen(uri)
            jdict = json.load(data)
            for site in jdict["features"]:
                lat = site["geometry"]["coordinates"][1]
                lon = site["geometry"]["coordinates"][0]
                if site["properties"]["sid"] == station:
                    station_metadata_dict = {}
                    station_metadata_dict["site_latitude"] = lat
                    station_metadata_dict["site_longitude"] = lon
                    for my_keys in site["properties"]:
                        if my_keys == "elevation":
                            station_metadata_dict["elevation"] = \
                                '%f meter' % site["properties"][my_keys]
                        else:
                            station_metadata_dict[my_keys] = \
                                site["properties"][my_keys]
                    metadata_list[station] = station_metadata_dict

        # Get station metadata
    else:
        raise ValueError("Either both lat_range and lon_range or station must " +
                         "be specified!")

    # Get the timestamp for each request
    start_time = time_window[0]
    end_time = time_window[1]

    SERVICE = "http://mesonet.agron.iastate.edu/cgi-bin/request/asos.py?"
    service = SERVICE + "data=all&tz=Etc/UTC&format=comma&latlon=yes&"

    service += start_time.strftime("year1=%Y&month1=%m&day1=%d&hour1=%H&minute1=%M&")
    service += end_time.strftime("year2=%Y&month2=%m&day2=%d&hour2=%H&minute2=%M")
    station_obs = {}
    for stations in site_list:
        uri = "%s&station=%s" % (service, stations)
        print("Downloading: %s" % (stations,))
        data = _download_data(uri)
        buf = StringIO()
        buf.write(data)
        buf.seek(0)

        my_df = pd.read_csv(buf, skiprows=5, na_values="M")

        if len(my_df['lat'].values) == 0:
            warnings.warn(
                "No data available at station %s between time %s and %s" %
                (stations, start_time.strftime('%Y-%m-%d %H:%M:%S'),
                 end_time.strftime('%Y-%m-%d %H:%M:%S')))
        else:
            def to_datetime(x):
                return datetime.strptime(x, "%Y-%m-%d %H:%M")

            my_df["time"] = my_df["valid"].apply(to_datetime)
            my_df = my_df.set_index("time")
            my_df = my_df.drop("valid", axis=1)
            my_df = my_df.drop("station", axis=1)
            my_df = my_df.to_xarray()

            my_df.attrs = metadata_list[stations]
            my_df["lon"].attrs["units"] = "degree"
            my_df["lon"].attrs["long_name"] = "Longitude"
            my_df["lat"].attrs["units"] = "degree"
            my_df["lat"].attrs["long_name"] = "Latitude"

            my_df["tmpf"].attrs["units"] = "degrees Fahrenheit"
            my_df["tmpf"].attrs["long_name"] = "Temperature in degrees Fahrenheit"

            # Fahrenheit to Celsius
            my_df["temp"] = (5. / 9. * my_df["tmpf"]) - 32.0
            my_df["temp"].attrs["units"] = "degrees Celsius"
            my_df["temp"].attrs["long_name"] = "Temperature in degrees Celsius"
            my_df["dwpf"].attrs["units"] = "degrees Fahrenheit"
            my_df["dwpf"].attrs["long_name"] = "Dewpoint temperature in degrees Fahrenheit"

            # Fahrenheit to Celsius
            my_df["dwpc"] = (5. / 9. * my_df["tmpf"]) - 32.0
            my_df["dwpc"].attrs["units"] = "degrees Celsius"
            my_df["dwpc"].attrs["long_name"] = "Dewpoint temperature in degrees Celsius"
            my_df["relh"].attrs["units"] = "percent"
            my_df["relh"].attrs["long_name"] = "Relative humidity"
            my_df["drct"].attrs["units"] = "degrees"
            my_df["drct"].attrs["long_name"] = "Wind speed in degrees"
            my_df["sknt"].attrs["units"] = "knots"
            my_df["sknt"].attrs["long_name"] = "Wind speed in knots"
            my_df["spdms"] = my_df["sknt"] * 0.514444
            my_df["spdms"].attrs["units"] = "m s-1"
            my_df["spdms"].attrs["long_name"] = "Wind speed in meters per second"
            my_df['u'] = -np.sin(np.deg2rad(my_df["drct"])) * my_df["spdms"]
            my_df['u'].attrs["units"] = "m s-1"
            my_df['u'].attrs["long_name"] = "Zonal component of surface wind"
            my_df['v'] = -np.cos(np.deg2rad(my_df["drct"])) * my_df["spdms"]
            my_df['v'].attrs["units"] = "m s-1"
            my_df['v'].attrs["long_name"] = "Meridional component of surface wind"
            my_df["mslp"].attrs["units"] = "mb"
            my_df["mslp"].attrs["long_name"] = "Mean Sea Level Pressure"
            my_df["alti"].attrs["units"] = "in Hg"
            my_df["alti"].attrs["long_name"] = "Atmospheric pressure in inches of Mercury"
            my_df["vsby"].attrs["units"] = "mi"
            my_df["vsby"].attrs["long_name"] = "Visibility"
            my_df["vsbykm"] = my_df["vsby"] * 1.60934
            my_df["vsbykm"].attrs["units"] = 'km'
            my_df["vsbykm"].attrs["long_name"] = "Visibility"
            my_df["gust"] = my_df["gust"] * 0.514444
            my_df["gust"].attrs["units"] = 'm s-1'
            my_df["gust"].attrs["long_name"] = "Wind gust speed"
            my_df["skyc1"].attrs["long_name"] = "Sky level 1 coverage"
            my_df["skyc2"].attrs["long_name"] = "Sky level 2 coverage"
            my_df["skyc3"].attrs["long_name"] = "Sky level 3 coverage"
            my_df["skyc4"].attrs["long_name"] = "Sky level 4 coverage"
            my_df["skyl1"] = my_df["skyl1"] * 0.3048
            my_df["skyl2"] = my_df["skyl2"] * 0.3048
            my_df["skyl3"] = my_df["skyl3"] * 0.3048
            my_df["skyl4"] = my_df["skyl4"] * 0.3048
            my_df["skyl1"].attrs["long_name"] = "Sky level 1 altitude"
            my_df["skyl2"].attrs["long_name"] = "Sky level 2 altitude"
            my_df["skyl3"].attrs["long_name"] = "Sky level 3 altitude"
            my_df["skyl4"].attrs["long_name"] = "Sky level 4 altitude"
            my_df["skyl1"].attrs["long_name"] = "meter"
            my_df["skyl2"].attrs["long_name"] = "meter"
            my_df["skyl3"].attrs["long_name"] = "meter"
            my_df["skyl4"].attrs["long_name"] = "meter"

            my_df["wxcodes"].attrs["long_name"] = "Weather code"
            my_df["ice_accretion_1hr"] = my_df["ice_accretion_1hr"] * 2.54
            my_df["ice_accretion_1hr"].attrs["units"] = "cm"
            my_df["ice_accretion_1hr"].attrs["long_name"] = "1 hour ice accretion"
            my_df["ice_accretion_3hr"] = my_df["ice_accretion_3hr"] * 2.54
            my_df["ice_accretion_3hr"].attrs["units"] = "cm"
            my_df["ice_accretion_3hr"].attrs["long_name"] = "3 hour ice accretion"
            my_df["ice_accretion_6hr"] = my_df["ice_accretion_3hr"] * 2.54
            my_df["ice_accretion_6hr"].attrs["units"] = "cm"
            my_df["ice_accretion_6hr"].attrs["long_name"] = "6 hour ice accretion"
            my_df["peak_wind_gust"] = my_df["peak_wind_gust"] * 0.514444
            my_df["peak_wind_gust"].attrs["units"] = 'm s-1'
            my_df["peak_wind_gust"].attrs["long_name"] = "Peak wind gust speed"
            my_df["peak_wind_drct"].attrs["drct"] = 'degree'
            my_df["peak_wind_drct"].attrs["long_name"] = "Peak wind gust direction"
            my_df['u_peak'] = -np.sin(np.deg2rad(my_df["peak_wind_drct"])) * my_df["peak_wind_gust"]
            my_df['u_peak'].attrs["units"] = "m s-1"
            my_df['u_peak'].attrs["long_name"] = "Zonal component of surface wind"
            my_df['v_peak'] = -np.cos(np.deg2rad(my_df["peak_wind_drct"])) * my_df["peak_wind_gust"]
            my_df['v_peak'].attrs["units"] = "m s-1"
            my_df['v_peak'].attrs["long_name"] = "Meridional component of surface wind"
            my_df["metar"].attrs["long_name"] = "Raw METAR code"
            my_df.attrs['_datastream'] = stations
            buf.close()

            station_obs[stations] = my_df
    return station_obs


def _download_data(uri):
    attempt = 0
    while attempt < 6:
        try:
            data = urlopen(uri, timeout=300).read().decode("utf-8")
            if data is not None and not data.startswith("ERROR"):
                return data
        except Exception as exp:
            print("download_data(%s) failed with %s" % (uri, exp))
            time.sleep(5)
        attempt += 1

    print("Exhausted attempts to download, returning empty data")
    return ""



""" Retrievals for precipitable water vapor. """

import numpy as np


def calculate_precipitable_water(ds, temp_name='tdry', rh_name='rh',
                                 pres_name='pres'):
    """

    Function to calculate precipitable water vapor from ARM sondewnpn b1 data.
    Will first calculate saturation vapor pressure of all data using Arden-Buck
    equations, then calculate specific humidity and integrate over all pressure
    levels to give us a precipitable water value in centimeters.

    ds : ACT object
        Object as read in by the ACT netCDF reader.
    temp_name : str
        Name of temperature field to use. Defaults to 'tdry' for sondewnpn b1
        level data.
    rh_name : str
        Name of relative humidity field to use. Defaults to 'rh' for sondewnpn
        b1 level data.
    pres_name : str
        Name of atmospheric pressure field to use. Defaults to 'pres' for
        sondewnpn b1 level data.

    """
    temp = ds[temp_name].values
    rh = ds[rh_name].values
    pres = ds[pres_name].values

    # Get list of temperature values for saturation vapor pressure calc
    temperature = []
    for t in np.nditer(temp):
        temperature.append(t)

    # Apply Arden-Buck equation to get saturation vapor pressure
    sat_vap_pres = []
    for t in temperature:
        # Over liquid water, above freezing
        if t >= 0:
            sat_vap_pres.append(0.61121 * np.exp((18.678 - (t / 234.5)) *
                                (t / (257.14 + t))))
        # Over ice, below freezing
        else:
            sat_vap_pres.append(0.61115 * np.exp((23.036 - (t / 333.7)) *
                                (t / (279.82 + t))))

    # convert rh from % to decimal
    rel_hum = []
    for r in np.nditer(rh):
        rel_hum.append(r / 100.)

    # get vapor pressure from rh and saturation vapor pressure
    vap_pres = []
    for i in range(0, len(sat_vap_pres)):
        es = rel_hum[i] * sat_vap_pres[i]
        vap_pres.append(es)

    # Get list of pressure values for mixing ratio calc
    pressure = []
    for p in np.nditer(pres):
        pressure.append(p)

    # Mixing ratio calc

    mix_rat = []
    for i in range(0, len(vap_pres)):
        mix_rat.append(0.622 * vap_pres[i] / (pressure[i] - vap_pres[i]))

    # Specific humidity

    spec_hum = []
    for rat in mix_rat:
        spec_hum.append(rat / (1 + rat))

    # Integrate specific humidity

    pwv = 0.0
    for i in range(1, len(pressure) - 1):
        pwv = pwv + 0.5 * (spec_hum[i] + spec_hum[i - 1]) * (pressure[i - 1] -
                                                             pressure[i])

    pwv = pwv / 0.098
    return pwv



import act
import numpy as np
import xarray as xr


def test_correct_ceil():
    # Make a fake ARM dataset to test with, just an array with 1e-7 for half
    # of it
    fake_data = 10 * np.ones((300, 20))
    fake_data[:, 10:] = -1
    arm_obj = {}
    arm_obj['backscatter'] = xr.DataArray(fake_data)
    arm_obj = act.corrections.ceil.correct_ceil(arm_obj)
    assert np.all(arm_obj['backscatter'].data[:, 10:] == -7)
    assert np.all(arm_obj['backscatter'].data[:, 1:10] == 1)


def test_correct_mpl():
    # Make a fake ARM dataset to test with, just an array with 1e-7 for half
    # of it
    test_data = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_MPL_1SAMPLE)
    obj = act.corrections.mpl.correct_mpl(test_data)
    sig_cross_pol = obj['signal_return_cross_pol'].values[1, 10:15]
    sig_co_pol = obj['signal_return_co_pol'].values[1, 10:15]
    height = obj['height'].values[0:10]
    overlap0 = obj['overlap_correction'].values[1, 0, 0:5]
    overlap1 = obj['overlap_correction'].values[1, 1, 0:5]
    overlap2 = obj['overlap_correction'].values[1, 2, 0:5]
    np.testing.assert_allclose(
        overlap0, [0., 0., 0., 0., 0.])
    np.testing.assert_allclose(
        overlap1, [754.338, 754.338, 754.338, 754.338, 754.338])
    np.testing.assert_allclose(
        overlap2, [181.9355, 181.9355, 181.9355, 181.9355,
                   181.9355])
    np.testing.assert_allclose(
        sig_cross_pol, [-0.5823283, -1.6066532, -1.7153032,
                        -2.520143, -2.275405], rtol=4e-07)
    np.testing.assert_allclose(
        sig_co_pol, [12.5631485, 11.035495, 11.999875,
                     11.09393, 11.388968])
    np.testing.assert_allclose(
        height, [0.00749012, 0.02247084, 0.03745109,
                 0.05243181, 0.06741206, 0.08239277, 0.09737302,
                 0.11235374, 0.12733398, 0.14231472], rtol=1e-6)
    assert obj['signal_return_co_pol'].attrs['units'] == '10 * log10(count/us)'
    assert obj['signal_return_cross_pol'].attrs['units'] == '10 * log10(count/us)'
    assert obj['cross_co_ratio'].attrs['long_name'] == 'Cross-pol / Co-pol ratio * 100'
    assert obj['cross_co_ratio'].attrs['units'] == 'LDR'
    assert 'description' not in obj['cross_co_ratio'].attrs.keys()
    assert 'ancillary_variables' not in obj['cross_co_ratio'].attrs.keys()
    assert np.all(np.round(obj['cross_co_ratio'].data[0, 500]) == 34.)
    assert np.all(np.round(obj['signal_return_co_pol'].data[0, 11]) == 11)
    assert np.all(np.round(obj['signal_return_co_pol'].data[0, 500]) == -6)
    test_data.close()
    obj.close()


def test_correct_wind():
    nav = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_NAV)
    nav = act.utils.ship_utils.calc_cog_sog(nav)

    aosmet = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_AOSMET)

    obj = xr.merge([nav, aosmet], compat='override')
    obj = act.corrections.ship.correct_wind(obj)

    assert round(obj['wind_speed_corrected'].values[800]) == 5.0
    assert round(obj['wind_direction_corrected'].values[800]) == 92.0


def test_correct_dl():
    # Test the DL correction script on a PPI dataset eventhough it will
    # mostlikely be used on FPT scans. Doing this to save space with only
    # one datafile in the repo.
    files = act.tests.sample_files.EXAMPLE_DLPPI
    obj = act.io.armfiles.read_netcdf(files)

    new_obj = act.corrections.doppler_lidar.correct_dl(obj, fill_value=np.nan)
    data = new_obj['attenuated_backscatter'].data
    data[np.isnan(data)] = 0.
    data = data * 100.
    data = data.astype(np.int64)
    assert np.sum(data) == -18633551

    new_obj = act.corrections.doppler_lidar.correct_dl(obj, range_normalize=False)
    data = new_obj['attenuated_backscatter'].data
    data[np.isnan(data)] = 0.
    data = data.astype(np.int64)
    assert np.sum(data) == -224000


def test_correct_rl():
    # Using ceil data in RL place to save memory
    files = act.tests.sample_files.EXAMPLE_RL1
    obj = act.io.armfiles.read_netcdf(files)

    obj = act.corrections.raman_lidar.correct_rl(obj,
                                                 range_normalize_log_values=True)
    np.testing.assert_almost_equal(np.max(obj['depolarization_counts_high'].values),
                                   9.91, decimal=2)
    np.testing.assert_almost_equal(np.min(obj['depolarization_counts_high'].values),
                                   -7.00, decimal=2)
    np.testing.assert_almost_equal(np.mean(obj['depolarization_counts_high'].values),
                                   -1.45, decimal=2)



"""
act.qc.qctests
------------------------------

Here we define the methods for performing the tests and putting the
results in the ancillary quality control varible. If you add a test
to this file you will need to add a method reference in the main
qcfilter class definition to make it callable.

"""

import numpy as np
import pandas as pd
import xarray as xr
import warnings
from act.utils import get_missing_value, convert_units


# This is a Mixins class used to allow using qcfilter class that is already
# registered to the xarray object. All the methods in this class will be added
# to the qcfilter class. Doing this to make the code spread across more files
# so it is more manageable and readable. Additinal files of tests can be added
# to qcfilter by creating a new class in the new file and adding to qcfilter
# class declaration.
class QCTests:
    """
    This is a Mixins class used to allow using qcfilter class that is already
    registered to the xarray object. All the methods in this class will be added
    to the qcfilter class. Doing this to make the code spread across more files
    so it is more manageable and readable. Additinal files of tests can be added
    to qcfilter by creating a new class in the new file and adding to qcfilter
    class declaration.

    """
    def __init__(self, obj, **kwargs):
        self._obj = obj

    def add_missing_value_test(self, var_name, missing_value=None,
                               missing_value_att_name='missing_value',
                               test_number=None, test_assessment='Bad',
                               test_meaning=None, flag_value=False,
                               prepend_text=None):
        """
        Method to add indication in quality control variable
        where data value is set to missing value.

        Parameters
        ----------
        var_name : str
            Data variable name.
        missing_value : int or float
            Optional missing value to use. If not provided will attempt
            to get it from the variable attribute or use NaN.
        missing_value_att_name : str
            Optional attribute name to use.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if test_meaning is None:
            test_meaning = 'Value is set to missing_value.'

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        if missing_value is None:
            missing_value = get_missing_value(self._obj, var_name, nodefault=True)
            if (missing_value is None and
                    self._obj[var_name].values.dtype.type in
                    (type(0.0), np.float16, np.float32, np.float64)):
                missing_value = float('nan')
            else:
                missing_value = -9999

        # Ensure missing_value attribute is matching data type
        missing_value = np.array(missing_value, dtype=self._obj[var_name].values.dtype.type)

        # New method using straight numpy instead of masked array
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            if np.isnan(missing_value) is False:
                index = np.equal(self._obj[var_name].values, missing_value)
            else:
                index = np.isnan(self._obj[var_name].values)

        test_dict = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        try:
            self._obj[var_name].attrs[missing_value_att_name]
        except KeyError:
            self._obj[var_name].attrs[missing_value_att_name] = missing_value

        return test_dict

    def add_less_test(self, var_name, limit_value, test_meaning=None,
                      test_assessment='Bad', test_number=None,
                      flag_value=False, limit_attr_name=None,
                      prepend_text=None):
        """
        Method to perform a less than test (i.e. minimum value) and add
        result to ancillary quality control variable. If ancillary
        quality control variable does not exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None, will return without adding test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_min'
            else:
                attr_name = 'fail_min'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = ('Data value less than {}.').format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        # New method with straight numpy
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.less(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_greater_test(self, var_name, limit_value, test_meaning=None,
                         test_assessment='Bad', test_number=None,
                         flag_value=False, limit_attr_name=None,
                         prepend_text=None):
        """
        Method to perform a greater than test (i.e. maximum value) and add
        result to ancillary quality control variable. If ancillary
        quality control variable does not exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None will return without setting test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_max'
            else:
                attr_name = 'fail_max'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = ('Data value greater than {}.').format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.greater(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_less_equal_test(self, var_name, limit_value, test_meaning=None,
                            test_assessment='Bad', test_number=None,
                            flag_value=False, limit_attr_name=None,
                            prepend_text=None):
        """
        Method to perform a less than or equal to test
        (i.e. minimum value) and add result to ancillary quality control
        variable. If ancillary quality control variable does not exist it
        will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None will return without setttin test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_min'
            else:
                attr_name = 'fail_min'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = ('Data value less than '
                            'or equal to {}.').format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.less_equal(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_greater_equal_test(self, var_name, limit_value, test_meaning=None,
                               test_assessment='Bad', test_number=None,
                               flag_value=False, limit_attr_name=None,
                               prepend_text=None):
        """
        Method to perform a greater than or equal to test
        (i.e. maximum value) and add result to ancillary quality control
        variable. If ancillary quality control variable does not exist it
        will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None will return without setttin test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_max'
            else:
                attr_name = 'fail_max'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = ('Data value greater than '
                            'or equal to {}.').format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.greater_equal(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_equal_to_test(self, var_name, limit_value, test_meaning=None,
                          test_assessment='Bad', test_number=None,
                          flag_value=False, limit_attr_name=None,
                          prepend_text=None):
        """
        Method to perform an equal test and add result to ancillary quality
        control variable. If ancillary quality control variable does not
        exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None will return without setttin test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_equal_to'
            else:
                attr_name = 'fail_equal_to'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = 'Data value equal to {}.'.format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.equal(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_not_equal_to_test(self, var_name, limit_value, test_meaning=None,
                              test_assessment='Bad', test_number=None,
                              flag_value=False, limit_attr_name=None,
                              prepend_text=None):
        """
        Method to perform a not equal to test and add result to ancillary
        quality control variable. If ancillary quality control variable does
        not exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value : int or float or None
            Limit value to use in test. The value will be written
            to the quality control variable as an attribute. If set
            to None will return without setttin test.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if limit_value is None:
            return

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_not_equal_to'
            else:
                attr_name = 'fail_not_equal_to'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = 'Data value not equal to {}.'.format(attr_name)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            index = np.not_equal(self._obj[var_name].values, limit_value)

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value = np.array(limit_value, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = limit_value

        return result

    def add_outside_test(self, var_name, limit_value_lower, limit_value_upper,
                         test_meaning=None,
                         test_assessment='Bad', test_number=None,
                         flag_value=False, limit_attr_names=None,
                         prepend_text=None):
        """
        Method to perform a less than or greater than test
        (i.e. outide minimum and maximum value) and add
        result to ancillary quality control variable. If ancillary
        quality control variable does not exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value_lower : int or float
            Lower limit value to use in test. The value will be written
            to the quality control variable as an attribute.
        limit_value_upper : int or float
            Upper limit value to use in test. The value will be written
            to the quality control variable as an attribute.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_names : list of str
            Optional attribute name to store the limit_value under
            quality control ancillary variable. First value is
            lower limit attribute name and second value is
            upper limit attribute name.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """

        if limit_attr_names is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name_lower = 'warn_lower_range'
                attr_name_upper = 'warn_upper_range'
            else:
                attr_name_lower = 'fail_lower_range'
                attr_name_upper = 'fail_upper_range'
        else:
            attr_name_lower = limit_attr_names[0]
            attr_name_upper = limit_attr_names[1]

        if test_meaning is None:
            test_meaning = ('Data value less than {} '
                            'or greater than {}.').format(attr_name_lower,
                                                          attr_name_upper)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            data = np.ma.masked_outside(self._obj[var_name].values,
                                        limit_value_lower, limit_value_upper)
        if data.mask.size == 1:
            data.mask = np.full(data.data.shape, data.mask, dtype=bool)

        index = data.mask

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value_lower = np.array(limit_value_lower, dtype=self._obj[var_name].values.dtype.type)
        limit_value_upper = np.array(limit_value_upper, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name_lower] = limit_value_lower
        self._obj[qc_var_name].attrs[attr_name_upper] = limit_value_upper

        return result

    def add_inside_test(self, var_name, limit_value_lower, limit_value_upper,
                        test_meaning=None, test_assessment='Bad',
                        test_number=None, flag_value=False,
                        limit_attr_names=None,
                        prepend_text=None):
        """
        Method to perform a greater than or less than test
        (i.e. between minimum and maximum value) and add
        result to ancillary quality control variable. If ancillary
        quality control variable does not exist it will be created.

        Parameters
        ----------
        var_name : str
            Data variable name.
        limit_value_lower : int or float
            Lower limit value to use in test. The value will be written
            to the quality control variable as an attribute.
        limit_value_upper : int or float
            Upper limit value to use in test. The value will be written
            to the quality control variable as an attribute.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        limit_attr_names : list of str
            Optional attribute name to store the limit_value under
            quality control ancillary variable. First value is
            lower limit attribute name and second value is
            upper limit attribute name.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """

        if limit_attr_names is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name_lower = 'warn_lower_range_inner'
                attr_name_upper = 'warn_upper_range_inner'
            else:
                attr_name_lower = 'fail_lower_range_inner'
                attr_name_upper = 'fail_upper_range_inner'
        else:
            attr_name_lower = limit_attr_names[0]
            attr_name_upper = limit_attr_names[1]

        if test_meaning is None:
            test_meaning = ('Data value greater than {} '
                            'or less than {}.').format(attr_name_lower,
                                                       attr_name_upper)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            data = np.ma.masked_inside(self._obj[var_name].values,
                                       limit_value_lower, limit_value_upper)
        if data.mask.size == 1:
            data.mask = np.full(data.data.shape, data.mask, dtype=bool)

        index = data.mask

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        # Ensure limit_value attribute is matching data type
        limit_value_lower = np.array(limit_value_lower, dtype=self._obj[var_name].values.dtype.type)
        limit_value_upper = np.array(limit_value_upper, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name_lower] = limit_value_lower
        self._obj[qc_var_name].attrs[attr_name_upper] = limit_value_upper

        return result

    def add_persistence_test(self, var_name, window=10, test_limit=0.0001,
                             min_periods=1, center=True, test_meaning=None,
                             test_assessment='Bad', test_number=None,
                             flag_value=False, prepend_text=None):
        """
        Method to perform a persistence test over 1-D data..

        Parameters
        ----------
        var_name : str
            Data variable name.
        window : int
            Optional number of data samples to use in the calculation of
            standard deviation to test for consistent data.
        test_limit : float
            Optional test limit to use where the standard deviation less
            than will trigger the test.
        min_periods : int
            Optional number of minimum values to use in the moving window.
            Setting to 1 so this correctly handles NaNs.
        center : boolean
            Optional where within the moving window to report the standard
            deviation values. Used in the .rolling.std() calculation with xarray.
        test_meaning : str
            The optional text description to add to flag_meanings
            describing the test. Will add a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will set a default if not set.
        test_number : int
            Optional test number to use. If not set will ues next
            available test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        data = self._obj[var_name]
        if window > data.size:
            window = data.size

        if test_meaning is None:
            test_meaning = ('Data failing persistence test. '
                            'Standard Deviation over a window of {} values '
                            'less than {}.').format(window, test_limit)

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            stddev = data.rolling(time=window, min_periods=min_periods, center=True).std()
            index = stddev < test_limit

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        return result

    def add_difference_test(self, var_name, dataset2_dict=None, ds2_var_name=None,
                            diff_limit=None, tolerance="1m",
                            set_test_regardless=True,
                            apply_assessment_to_dataset2=None,
                            apply_tests_to_dataset2=None,
                            test_meaning=None, test_assessment='Bad',
                            test_number=None, flag_value=False,
                            prepend_text=None):
        """
        Method to perform a comparison test on time series data. Tested on 1-D
        data only. Will check if units and long_name indicate a direction and
        compensate for 0 to 360 degree transition.

        Parameters
        ----------
        var_name : str
            Data variable name.
        dataset2_dict : dict
            Dictionary with key equal to datastream name and value
            equal to xarray dataset containging variable to compare. If no provided
            will assume second dataset is the same as self dataset.
        ds2_var_name : str
            Comparison dataset variable name to compare.
        diff_limit : int or float
            Difference limit for comparison.
        apply_assessment_to_dataset2 : str or list of str
            Option to filter comparison dataset variable using corresponsing
            quality control variable using assessments. Example would be
            ['Bad'], where all quality control data with assessment Bad will
            not be used in this test.
        apply_tests_to_dataset2 : int or list of int
            Option to filter comparison dataset variable using corresponding
            quality control variable using test numbers. Example would be
            [2,4], where all quality control data with test numbers 2 or 4 set
            will not be used in this test.
        tolerance : str
            Optional text indicating the time tolerance for aligning two
            DataArrays.
        set_test_regardless : boolean
            Option to set test description even if no data in comparison data
            set.
        test_meaning : str
            Optional text description to add to flag_meanings
            describing the test. Will use a default if not set.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will use a default if not set.
        test_number : int
            Optional test number to use. If not set will use next available
            test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """
        if dataset2_dict is None:
            dataset2_dict = {'second_dataset': self._obj}

        if not isinstance(dataset2_dict, dict):
            raise ValueError('You did not provide a dictionary containing the '
                             'datastream name as the key and xarray dataset as '
                             'the value for dataset2_dict for add_difference_test().')

        if diff_limit is None:
            raise ValueError('You did not provide a test limit for add_difference_test().')

        datastream2 = list(dataset2_dict.keys())[0]
        dataset2 = dataset2_dict[datastream2]

        if set_test_regardless is False and type(dataset2) != xr.core.dataset.Dataset:
            return

        if test_meaning is None:
            if dataset2 is self._obj:
                var_name2 = f'{ds2_var_name}'
            else:
                var_name2 = f'{datastream2}:{ds2_var_name}'

            test_meaning = (f'Difference between {var_name} and {var_name2} '
                            f'greater than {diff_limit} {self._obj[var_name].attrs["units"]}')

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        if tolerance is not None:
            tolerance = pd.Timedelta(tolerance)

        index = []
        if type(dataset2) == xr.core.dataset.Dataset:
            if apply_assessment_to_dataset2 is not None or apply_tests_to_dataset2 is not None:
                dataset2[ds2_var_name].values = dataset2.qcfilter.get_masked_data(
                    ds2_var_name, rm_assessments=apply_assessment_to_dataset2,
                    rm_tests=apply_tests_to_dataset2, return_nan_array=True)

            df_a = pd.DataFrame({'time': self._obj['time'].values,
                                 var_name: self._obj[var_name].values})
            data_b = convert_units(dataset2[ds2_var_name].values,
                                   dataset2[ds2_var_name].attrs['units'],
                                   self._obj[var_name].attrs['units'])
            ds2_var_name = ds2_var_name + '_newname'
            df_b = pd.DataFrame({'time': dataset2['time'].values,
                                 ds2_var_name: data_b})

            if tolerance is not None:
                tolerance = pd.Timedelta(tolerance)

            pd_c = pd.merge_asof(df_a, df_b, on='time', tolerance=tolerance,
                                 direction="nearest")

            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                # Check if variable is for wind direction comparisons. Fix
                # for 0 - 360 degrees transition. This is done by adding 360 degrees to
                # all wind values and using modulus to get the minimum difference number.
                # This is done for both a-b and b-a and then choosing the minimum number
                # to compensate for large differences.
                wdir_units = ['deg', 'degree', 'degrees', 'degs']
                if (self._obj[var_name].attrs['units'] in wdir_units and
                        'direction' in self._obj[var_name].attrs['long_name'].lower()):
                    diff1 = np.mod(np.absolute((pd_c[var_name] + 360.) -
                                   (pd_c[ds2_var_name] + 360.)), 360)
                    diff2 = np.mod(np.absolute((pd_c[ds2_var_name] + 360.) -
                                   (pd_c[var_name] + 360.)), 360)
                    diff = np.array([diff1, diff2])
                    diff = np.nanmin(diff, axis=0)

                else:
                    diff = np.absolute(pd_c[var_name] - pd_c[ds2_var_name])

                index = diff > diff_limit

        result = self._obj.qcfilter.add_test(
            var_name, index=index,
            test_number=test_number,
            test_meaning=test_meaning,
            test_assessment=test_assessment,
            flag_value=flag_value)

        return result

    def add_delta_test(self, var_name, diff_limit=1, test_meaning=None,
                       limit_attr_name=None,
                       test_assessment='Indeterminate', test_number=None,
                       flag_value=False, prepend_text=None):
        """
        Method to perform a difference test on adjacent values in time series.
        Will flag both values where a difference is greater
        than or equal to the difference limit. Tested with 1-D data only. Not
        sure what will happen with higher dimentioned data.

        Parameters
        ----------
        var_name : str
            Data variable name.
        diff_limit : int or float
            Difference limit
        test_meaning : str
            Optional text description to add to flag_meanings
            describing the test. Will use a default if not set.
        limit_attr_name : str
            Optional attribute name to store the limit_value under
            quality control ancillary variable.
        test_assessment : str
            Optional single word describing the assessment of the test.
            Will use a default if not set.
        test_number : int
            Optional test number to use. If not set will use next available
            test number.
        flag_value : boolean
            Indicates that the tests are stored as integers
            not bit packed values in quality control variable.
        prepend_text : str
            Optional text to prepend to the test meaning.
            Example is indicate what institution added the test.

        Returns
        -------
        test_info : tuple
            A tuple containing test information including var_name, qc variable name,
            test_number, test_meaning, test_assessment

        """

        if limit_attr_name is None:
            if test_assessment == 'Suspect' or test_assessment == 'Indeterminate':
                attr_name = 'warn_delta'
            else:
                attr_name = 'fail_delta'
        else:
            attr_name = limit_attr_name

        if test_meaning is None:
            test_meaning = f'Difference between current and previous values exceeds {attr_name}.'

        if prepend_text is not None:
            test_meaning = ': '.join((prepend_text, test_meaning))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            # Check if variable is for wind direction comparisons by units. Fix
            # for 0 - 360 degrees transition. This is done by adding 360 degrees to
            # all wind values and using modulus to get the minimum difference number.
            wdir_units = ['deg', 'degree', 'degrees', 'degs']
            if (self._obj[var_name].attrs['units'] in wdir_units and
                    'direction' in self._obj[var_name].attrs['long_name'].lower()):
                abs_diff = np.mod(np.abs(np.diff(self._obj[var_name].values)), 360)
            else:
                abs_diff = np.abs(np.diff(self._obj[var_name].values))

            index = np.where(abs_diff >= diff_limit)[0]
            if index.size > 0:
                index = np.append(index, index + 1)
                index = np.unique(index)

        result = self._obj.qcfilter.add_test(var_name, index=index,
                                             test_number=test_number,
                                             test_meaning=test_meaning,
                                             test_assessment=test_assessment,
                                             flag_value=flag_value)

        # Ensure min value attribute is matching data type
        diff_limit = np.array(diff_limit, dtype=self._obj[var_name].values.dtype.type)

        qc_var_name = result['qc_variable_name']
        self._obj[qc_var_name].attrs[attr_name] = diff_limit

        return result



"""
=====================
act.utils (act.utils)
=====================

.. currentmodule:: act.utils

This module contains the common procedures used by all modules of the ARM
Community Toolkit.

.. autosummary::
    :toctree: generated/

    accumulate_precip
    add_in_nan
    add_solar_variable
    assign_coordinates
    calc_cog_sog
    calculate_dqr_times
    convert_units
    create_pyart_obj
    dates_between
    datetime64_to_datetime
    decode_present_weather
    destination_azimuth_distance
    determine_time_delta
    get_missing_value
    numpy_to_arm_date
    planck_converter
    reduce_time_ranges
    ts_weighted_average
"""

from .data_utils import add_in_nan
from .data_utils import get_missing_value
from .data_utils import convert_units
from .data_utils import assign_coordinates
from .data_utils import accumulate_precip
from .data_utils import ts_weighted_average
from .data_utils import create_pyart_obj
from .datetime_utils import dates_between
from .datetime_utils import numpy_to_arm_date
from .datetime_utils import reduce_time_ranges
from .datetime_utils import determine_time_delta
from .datetime_utils import datetime64_to_datetime
from .qc_utils import calculate_dqr_times
from .ship_utils import calc_cog_sog
from .geo_utils import destination_azimuth_distance
from .geo_utils import add_solar_variable
from .inst_utils import decode_present_weather
from .radiance_utils import planck_converter
from .data_utils import ChangeUnits



"""
=================================
act.corrections (act.corrections)
=================================

.. currentmodule:: act.corrections

The procedures in this module contain corrections for various datasets.

.. autosummary::
    :toctree: generated/

    ceil.correct_ceil
    doppler_lidar.correct_dl
    mpl.correct_mpl
    raman_lidar.correct_rl
    ship.correct_wind
"""

from . import ceil
from . import mpl
from . import ship
from . import doppler_lidar
from . import raman_lidar



"""
===========================
act.qc (act.qc)
===========================

.. currentmodule:: act.qc

This module contains procedures for working with QC information
and for applying tests to data

.. autosummary::
    :toctree: generated/

    arm.add_dqr_to_qc
    clean.CleanDataset
    qcfilter.QCFilter
    qcfilter.parse_bit
    qcfilter.set_bit
    qcfilter.unset_bit
    qcfilter.QCFilter.add_delta_test
    qcfilter.QCFilter.add_difference_test
    qcfilter.QCFilter.add_equal_to_test
    qcfilter.QCFilter.add_greater_test
    qcfilter.QCFilter.add_greater_equal_test
    qcfilter.QCFilter.add_inside_test
    qcfilter.QCFilter.add_less_test
    qcfilter.QCFilter.add_less_equal_test
    qcfilter.QCFilter.add_missing_value_test
    qcfilter.QCFilter.add_not_equal_to_test
    qcfilter.QCFilter.add_outside_test
    qcfilter.QCFilter.add_persistence_test
    qcfilter.QCFilter.add_test
    qcfilter.QCFilter.available_bit
    qcfilter.QCFilter.check_for_ancillary_qc
    qcfilter.QCFilter.compare_time_series_trends
    qcfilter.QCFilter.create_qc_variable
    qcfilter.QCFilter.get_qc_test_mask
    qcfilter.QCFilter.get_masked_data
    qcfilter.QCFilter.remove_test
    qcfilter.QCFilter.set_test
    qcfilter.QCFilter.unset_test
    qcfilter.QCFilter.update_ancillary_variable
    qctests.QCTests
    radiometer_tests.fft_shading_test
"""

from . import qcfilter
from . import qctests, comparison_tests
from . import clean
from . import arm
from . import radiometer_tests



"""
===============
act.io (act.io)
===============

.. currentmodule:: act.io

This module contains procedures for reading and writing various ARM datasets.

.. autosummary::
    :toctree: generated/

    armfiles.check_arm_standards
    armfiles.create_obj_from_arm_dod
    armfiles.read_netcdf
    csvfiles.read_csv
    mpl.read_sigma_mplv5
"""

from . import armfiles
from . import csvfiles
from . import mpl
from . import noaagml



""" Functions for correcting ceilometer data. """

import numpy as np


def correct_ceil(obj, fill_value=1e-7, var_name='backscatter'):
    """
    This procedure corrects ceilometer data by filling all zero and
    negative values of backscatter with fill_value and then converting
    the backscatter data into logarithmic space.

    Parameters
    ----------
    obj : Dataset object
        The ceilometer dataset to correct. The backscatter data should be
        in linear space.
    var_name : str
        The variable name of data in the object.
    fill_value : float
        The fill_value to use. The fill_value is entered in linear space.

    Returns
    -------
    obj : Dataset object
        The ceilometer dataset containing the corrected values.

    """
    backscat = obj[var_name].data
    backscat[backscat <= 0] = fill_value
    backscat = np.log10(backscat)

    obj[var_name].data = backscat

    return obj



"""
act.plotting.plot
=================

Class for creating timeseries plots from ACT datasets.

"""

# Import third party libraries
import matplotlib.pyplot as plt
import numpy as np
import warnings
import xarray as xr


class Display(object):
    """
    This class is the base class for all of the other Display object
    types in ACT. This contains the common attributes and routines
    between the differing *Display* classes. We recommend that you
    use the classes inherited from Display for making your plots
    such as :func:`act.plotting.TimeSeriesDisplay` and
    :func:`act.plotting.WindRoseDisplay` instead of
    trying to do so using the Display object.

    However, we do ask that if you add another object to the plotting
    module of ACT that you make it a subclass of Display. Display provides
    some basic functionality for the handling of datasets and subplot
    parameters.

    Attributes
    ----------
    fields : dict
        The dictionary containing the fields inside the ARM dataset. Each field
        has a key that links to an xarray DataArray object.
    ds : str
        The name of the datastream.
    file_dates : list
        The dates of each file being displayed.
    fig : matplotlib figure handle
        The matplotlib figure handle to display the plots on. Initializing the
        class with this set to None will create a new figure handle. See the
        matplotlib documentation on what keyword arguments are
        available.
    axes : list
        The list of axes handles to each subplot.
    plot_vars : list
        The list of variables being plotted.
    cbs : list
        The list of colorbar handles.

    Parameters
    ----------
    obj : ACT Dataset, dict, or tuple
        The ACT Dataset to display in the object. If more than one dataset
        is to be specified, then a tuple can be used if all of the datasets
        conform to ARM standards. Otherwise, a dict with a key corresponding
        to the name of each datastream will need to be supplied in order
        to create the ability to plot multiple datasets.
    subplot_shape : 1 or 2D tuple
        A tuple representing the number of (rows, columns) for the subplots
        in the display. If this is None, the figure and axes will not
        be initialized.
    ds_name : str or None
        The name of the datastream to plot. This is only used if a non-ARM
        compliant dataset is being loaded and if only one such dataset is
        loaded.
    subplot_kw : dict, optional
        The kwargs to pass into :func:`fig.subplots`
    **kwargs : keywords arguments
        Keyword arguments passed to :func:`plt.figure`.

    """
    def __init__(self, obj, subplot_shape=(1,), ds_name=None,
                 subplot_kw=None, **kwargs):
        if isinstance(obj, xr.Dataset):
            if 'datastream' in obj.attrs.keys() is not None:
                self._arm = {obj.attrs['datastream']: obj}
            elif ds_name is not None:
                self._arm = {ds_name: obj}
            else:
                warnings.warn(("Could not discern datastream" +
                               "name and dict or tuple were " +
                               "not provided. Using default" +
                               "name of act_datastream!"), UserWarning)

                self._arm = {'act_datastream': obj}

        # Automatically name by datastream if a tuple of object is supplied
        if isinstance(obj, tuple):
            self._arm = {}
            for objs in obj:
                self._arm[objs.attrs['datastream']] = objs

        if isinstance(obj, dict):
            self._arm = obj

        self.fields = {}
        self.ds = {}
        self.file_dates = {}
        self.xrng = np.zeros((1, 2))
        self.yrng = np.zeros((1, 2))

        for dsname in self._arm.keys():
            self.fields[dsname] = self._arm[dsname].variables
            if '_datastream' in self._arm[dsname].attrs.keys():
                self.ds[dsname] = str(self._arm[dsname].attrs['_datastream'])
            else:
                self.ds[dsname] = "act_datastream"
            if '_file_dates' in self._arm[dsname].attrs.keys():
                self.file_dates[dsname] = self._arm[dsname].attrs['_file_dates']

        self.fig = None
        self.axes = None
        self.plot_vars = []
        self.cbs = []
        if subplot_shape is not None:
            self.add_subplots(subplot_shape, subplot_kw=subplot_kw,
                              **kwargs)

    def add_subplots(self, subplot_shape=(1,), subplot_kw=None,
                     **kwargs):
        """
        Adds subplots to the Display object. The current
        figure in the object will be deleted and overwritten.

        Parameters
        ----------
        subplot_shape : 1 or 2D tuple, list, or array
            The structure of the subplots in (rows, cols).
        subplot_kw : dict, optional
            The kwargs to pass into fig.subplots.
        **kwargs : keyword arguments
            Any other keyword arguments that will be passed
            into :func:`matplotlib.pyplot.subplots`. See the matplotlib
            documentation for further details on what keyword
            arguments are available.

        """
        if self.fig is not None:
            plt.close(self.fig)
            del self.fig

        if len(subplot_shape) == 2:
            fig, ax = plt.subplots(
                subplot_shape[0], subplot_shape[1],
                subplot_kw=subplot_kw,
                **kwargs)
            self.xrng = np.zeros((subplot_shape[0], subplot_shape[1], 2))
            self.yrng = np.zeros((subplot_shape[0], subplot_shape[1], 2))
            if subplot_shape[0] == 1:
                ax = ax.reshape(1, subplot_shape[1])
        elif len(subplot_shape) == 1:
            fig, ax = plt.subplots(
                subplot_shape[0], 1, subplot_kw=subplot_kw, **kwargs)
            if subplot_shape[0] == 1:
                ax = np.array([ax])
            self.xrng = np.zeros((subplot_shape[0], 2))
            self.yrng = np.zeros((subplot_shape[0], 2))
        else:
            raise ValueError(("subplot_shape must be a 1 or 2 dimensional" +
                              "tuple list, or array!"))
        self.fig = fig
        self.axes = ax

    def put_display_in_subplot(self, display, subplot_index):
        """
        This will place a Display object into a specific subplot.
        The display object must only have one subplot.

        This will clear the display in the Display object being added.

        Parameters
        ----------
        Display : Display object or subclass
            The Display object to add as a subplot
        subplot_index : tuple
            Which subplot to add the Display to.

        Returns
        -------
        ax : matplotlib axis handle
            The axis handle to the display object being added.

        """
        if len(display.axes) > 1:
            raise RuntimeError("Only single plots can be made as subplots " +
                               "of another Display object!")

        my_projection = display.axes[0].name
        plt.close(display.fig)
        display.fig = self.fig
        self.fig.delaxes(self.axes[subplot_index])
        the_shape = self.axes.shape
        if len(the_shape) == 1:
            second_value = 1
        else:
            second_value = the_shape[1]

        self.axes[subplot_index] = self.fig.add_subplot(
            the_shape[0], second_value,
            (second_value - 1) * the_shape[0] + subplot_index[0] + 1,
            projection=my_projection)

        display.axes = np.array([self.axes[subplot_index]])

        return display.axes[0]

    def assign_to_figure_axis(self, fig, ax):
        """
        This assigns the Display to a specific figure and axis.
        This will remove the figure and axes that are currently
        stored in the object. The display object will then only
        have one axis handle.

        Parameters
        ----------
        fig : matplotlib figure handle
            The figure to place the time series display in.
        ax : axis handle
            The axis handle to place the plot in.

        """
        if self.fig is not None:
            plt.close(self.fig)
            del self.fig

        del self.axes
        self.fig = fig
        self.axes = np.array([ax])

    def add_colorbar(self, mappable, title=None, subplot_index=(0, )):
        """
        Adds a colorbar to the plot.

        Parameters
        ----------
        mappable : matplotlib mappable
            The mappable to base the colorbar on.
        title : str
            The title of the colorbar. Set to None to have no title.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        Returns
        -------
        cbar : matplotlib colorbar handle
            The handle to the matplotlib colorbar.

        """
        if self.axes is None:
            raise RuntimeError("add_colorbar requires the plot "
                               "to be displayed.")

        fig = self.fig
        ax = self.axes[subplot_index]

        # Give the colorbar it's own axis so the 2D plots line up with 1D
        box = ax.get_position()
        pad, width = 0.01, 0.01
        cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])
        cbar = plt.colorbar(mappable, cax=cax)
        cbar.ax.set_ylabel(title, rotation=270, fontsize=8, labelpad=3)
        cbar.ax.tick_params(labelsize=6)
        self.cbs.append(cbar)

        return cbar



"""
act.qc.radiometer_tests
------------------------------

Tests specific to radiometers
"""

from scipy.fftpack import rfft, rfftfreq
import numpy as np
import xarray as xr
import pandas as pd
import datetime
import dask
import warnings

from act.utils.datetime_utils import determine_time_delta
from act.utils.geo_utils import get_sunrise_sunset_noon, is_sun_visible


def fft_shading_test(obj, variable='diffuse_hemisp_narrowband_filter4',
                     fft_window=30,
                     shad_freq_lower=[0.008, 0.017],
                     shad_freq_upper=[0.0105, 0.0195],
                     ratio_thresh=[3.15, 1.2],
                     time_interval=None, smooth_window=5, shading_thresh=0.4):
    """
    Function to test shadowband radiometer (MFRSR, RSS, etc) instruments
    for shading related problems.  Program was adapted by Adam Theisen
    from the method defined in Alexandrov et al 2007 to process on a
    point by point basis using a window of data around that point for
    the FFT analysis.

    For ARM data, testing has found that this works the best on narrowband
    filter4 for MFRSR data.

    Function has been tested and is in use by the ARM DQ Office for
    problem detection.  It is know to have some false positives at times.

    Need to run obj.clean.cleanup() ahead of time to ensure proper addition
    to QC variable

    Parameters
    ----------
    obj : xarray Dataset
        Data object
    variable : string
        Name of variable to process
    fft_window : int
        Number of samples to use in the FFT window.  Default is +- 30 samples
        Note: this is +- so the full window will be double
    shad_freq_lower : list
        Lower frequency over which to look for peaks in FFT
    shad_freq_upper : list
        Upper frequency over which to look for peaks in FFT
    ratio_thresh : list
        Threshold for each freq window to flag data.  I.e. if the peak is 3.15 times
        greater than the surrounding area
    time_interval : float
        Sampling rate of the instrument
    smooth_window : int
        Number of samples to use in smoothing FFTs before analysis
    shading_thresh : float
        After smoothing, the value over which is considered a shading signal

    Returns
    -------
    obj : xarray Dataset
        Data object

    References
    ----------
    Alexandrov, Mikhail & Kiedron, Peter & Michalsky, Joseph & Hodges, Gary
    & Flynn, Connor & Lacis, Andrew. (2007). Optical depth measurements by
    shadow-band radiometers and their uncertainties. Applied optics. 46.
    8027-38. 10.1364/AO.46.008027.

    """

    # Get time and data from variable
    time = obj['time'].values
    data = obj[variable].values
    if 'missing_value' in obj[variable].attrs:
        missing = obj[variable].attrs['missing_value']
    else:
        missing = -9999.

    # Get time interval between measurements
    if time_interval is None:
        dt = determine_time_delta(time)
    else:
        dt = time_interval

    # Compute the FFT for each point +- window samples
    task = []
    sun_up = is_sun_visible(latitude=obj['lat'].values, longitude=obj['lon'].values, date_time=time)
    for t in range(len(time)):
        sind = t - fft_window
        eind = t + fft_window
        if sind < 0:
            sind = 0
        if eind > len(time):
            eind = len(time)

        # Get data and remove all nan/missing values
        d = data[sind:eind]
        idx = ((d != missing) & (np.isnan(d) is not True))
        index = np.where(idx)
        d = d[index]

        # Add to task for dask processing
        task.append(dask.delayed(fft_shading_test_process)(
            time[t], d,
            shad_freq_lower=shad_freq_lower,
            shad_freq_upper=shad_freq_upper,
            ratio_thresh=ratio_thresh,
            time_interval=dt,
            is_sunny=sun_up[t]))

    # Process using dask
    result = dask.compute(*task)

    # Run data through a rolling median to filter out singular
    # false positives
    shading = [r['shading'] for r in result]
    shading = pd.Series(shading).rolling(window=smooth_window, min_periods=1).median()

    # Find indices where shading is indicated
    idx = (np.asarray(shading) > shading_thresh)
    index = np.where(idx)

    # Add test to QC Variable
    desc = 'FFT Shading Test'
    obj.qcfilter.add_test(variable, index=index, test_meaning=desc)

    # Prepare frequency and fft variables for adding to object
    fft = np.empty([len(time), fft_window * 2])
    fft[:] = np.nan
    freq = np.empty([len(time), fft_window * 2])
    freq[:] = np.nan
    for i, r in enumerate(result):
        dummy = r['fft']
        fft[i, 0:len(dummy)] = dummy
        dummy = r['freq']
        freq[i, 0:len(dummy)] = dummy

    attrs = {'units': '', 'long_name': 'FFT Results for Shading Test', 'upper_freq': shad_freq_upper,
             'lower_freq': shad_freq_lower}
    fft_window = xr.DataArray(range(fft_window * 2), dims=['fft_window'],
                              attrs={'long_name': 'FFT Window', 'units': '1'})
    da = xr.DataArray(fft, dims=['time', 'fft_window'], attrs=attrs, coords=[obj['time'], fft_window])
    obj['fft'] = da
    attrs = {'units': '', 'long_name': 'FFT Frequency Values for Shading Test'}
    da = xr.DataArray(freq, dims=['time', 'fft_window'], attrs=attrs, coords=[obj['time'], fft_window])
    obj['fft_freq'] = da

    return obj


def fft_shading_test_process(time, data, shad_freq_lower=None,
                             shad_freq_upper=None, ratio_thresh=None,
                             time_interval=None, is_sunny=None):
    """
    Processing function to do the FFT calculations/thresholding

    Parameters
    ----------
    time : datetime
        Center time of calculation used for calculating sunrise/sunset
    data : list
        Data for run through fft processing
    shad_freq_lower : list
        Lower limits of freqencies to look for shading issues
    shad_freq_upper : list
        Upper limits of freqencies to look for shading issues
    ratio_thresh : list
        Thresholds to apply, corresponding to frequencies chosen
    time_interval : float
        Time interval of data

    Returns
    -------
    shading : int
        Binary to indicate shading problem (1) or not (0)

    """

    if not is_sunny:
        return {'shading': 0, 'fft': [np.nan] * len(data), 'freq': [np.nan] * len(data)}

    # FFT Algorithm
    fftv = abs(rfft(data))
    freq = rfftfreq(fftv.size, d=time_interval)

    # Get FFT data under threshold
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        idx = (fftv > 1.)
    index = np.where(idx)
    fftv[index] = np.nan
    freq[index] = np.nan

    # Return if FFT is empty
    if len(fftv) == 0:
        return {'shading': 0, 'fft': [np.nan] * len(data), 'freq': [np.nan] * len(data)}

    # Commented out as it seems to work better without smoothing
    # fftv=pd.DataFrame(data=fftv).rolling(min_periods=3,window=3,center=True).mean().values.flatten()

    ratio = []
    # Calculates the ratio (size) of the peaks in the FFT to the surrounding
    # data
    wind = 3

    # Run through each frequency to look for peaks
    # Calculate threshold of peak value to surrounding values
    for i in range(len(shad_freq_lower)):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            idx = np.logical_and(freq > shad_freq_lower[i],
                                 freq < shad_freq_upper[i])

        index = np.where(idx)
        if len(index[0]) == 0:
            continue
        peak = max(fftv[index])
        index = index[0]

        sind = index[0] - wind
        if sind < 0:
            sind = 0
        eind = index[-1] + wind
        if eind > len(fftv):
            eind = len(fftv)

        if len(range(sind, index[0])) == 0 or len(range(index[-1], eind)) == 0:
            ratio.append(0.0)
        else:
            # Calculates to the left/right of each peak
            peak_l = max(fftv[range(sind, index[0])])
            peak_r = max(fftv[range(index[-1], eind)])
            ratio.append(peak / np.mean([peak_l, peak_r]))

    # Checks ratios against thresholds for each freq range
    shading = 0
    if len(ratio) > 0:
        pass1 = False
        pass2 = False
        if ratio[0] > ratio_thresh[0]:
            pass1 = True
        if len(ratio) > 1:
            if ratio[1] > ratio_thresh[1]:
                pass2 = True
        else:
            pass2 = True

        if pass1 and pass2:
            shading = 1

    return {'shading': shading, 'fft': fftv, 'freq': freq}



""" Functions for correcting wind speed and direction for ship motion. """

import numpy as np


def correct_wind(obj, wspd_name='wind_speed', wdir_name='wind_direction',
                 heading_name='yaw', cog_name='course_over_ground',
                 sog_name='speed_over_ground'):
    """
    This procedure corrects wind speed and direction for ship motion
    based on equations from NOAA tech. memo. PSD-311. A Guide to Making
    Climate Quality Meteorological and Flux Measurements at Sea.
    https://www.go-ship.org/Manual/fluxhandbook_NOAA-TECH%20PSD-311v3.pdf

    Parameters
    ----------
    obj : Dataset object
        The ceilometer dataset to correct. The backscatter data should be
        in linear space.
    wspd_name : string
        Wind speed variable name.
    wdir_name : string
        Wind direction variable name.
    heading_name : string
        Navigiation heading variable name.
    cog_name : string
        Course over ground variable name.
    sog_name : string
        Speed over ground variable name.

    Returns
    -------
    obj : Dataset object
        The dataset containing the corrected values.

    References
    ----------
    Bradley, F. and Farall. C. (2007) A Guide to Making Climate Quality Meteorological
    and Flux Measurements at Sea. Boulder, CO, NOAA, Earth System Research Laboratory,
    Physical Sciences Division, 44pp. & appendices. (NOAA Technical Memorandum OAR PSD-311).
    http://hdl.handle.net/11329/386

    """
    # Set variables to be used and convert to radians
    rels = obj[wspd_name]
    reld = np.deg2rad(obj[wdir_name])
    head = np.deg2rad(obj[heading_name])
    cog = np.deg2rad(obj[cog_name])
    sog = obj[sog_name]

    # Calculate winds based on method in the document denoted above
    relsn = rels * np.cos(head + reld)
    relse = rels * np.sin(head + reld)

    sogn = sog * np.cos(cog)
    soge = sog * np.sin(cog)

    un = relsn - sogn
    ue = relse - soge

    dirt = np.mod(np.rad2deg(np.arctan2(ue, un)) + 360., 360)
    ut = np.sqrt(un ** 2. + ue ** 2)

    # Create data arrays and add corrected wind direction and speed
    # to the initial object that was passed in
    wdir_da = obj[wdir_name].copy(data=dirt)
    wdir_da.attrs['long_name'] = 'Wind direction corrected to ship motion'
    obj[wdir_name + '_corrected'] = wdir_da

    wspd_da = obj[wspd_name].copy(data=ut)
    wspd_da.attrs['long_name'] = 'Wind speed corrected to ship motion'
    obj[wspd_name + '_corrected'] = wspd_da

    return obj



"""
act.utils.radiance_utils
--------------------

Module containing utilities for radiance calculations

"""


import numpy as np
import inspect


def planck_converter(wnum=None, radiance=None, temperature=None,
                     units='cm'):
    """
    Planck function to convert radiance to temperature or temperature to
    radiance given a corresponding wavenumber value.

    Constants have been updated to reflect those used on the NOAA site
    https://ncc.nesdis.noaa.gov/data/planck.html

    Parameters
    ----------
    wnum : float or list or numpy array
        Wave number for corresponding radiance or temperature value.
    radiance : float or list or numpy array
        Radiance value of corresponding wavenumber in W/m^2/sr/cm.
    temperature : float or list or numpy array
        Temperature value of corresponding wavenumber in degK/cm.
    units : string
        Units value to use.  Default to cm, other option is m

    References
    ----------
    This procedure was created for the AERI vs. IRT comparison from F77 code
    orginally written by Jim Liligren and passed to the DQO by the ARM IRT
    mentor.

    Example
    -------
    > planck_converter(wnum=1100, temperature=300)
    81.49372361846207

    > planck_converter(wnum=1100, radiance=81.49)
    299.9974150039702

    """
    # h = Plancks's constant
    # k = Boltzmann's constant
    # c = Speed of light in a vacuum
    if units == 'cm':
        # C1 = 2 h c^2, [W cm^2]
        C1 = 1.191042e-5  # For radiance in units of mW m^-2 sr^-1 /cm^-1
        # C2 = h c / k, [K cm]
        C2 = 1.4387752
    if units == 'm':
        # C1 = 2 h c^2, [W m^2]
        C1 = 1.191042e-16  # For radiance in units of mW m^-2 sr^-1 /cm^-1
        # C2 = h c / k, [K m]
        C2 = 1.4387752e-2

    func_name = inspect.stack()[0][3]
    if wnum is None:
        raise ValueError(f"No wnum values provided for {func_name}() function.\n")

    if radiance is None and temperature is None:
        raise ValueError(f"No radiance or temperature values provided for {func_name}() function.\n")

    if radiance is not None:
        radiance = np.array(radiance, dtype=np.float64)
        wnum = np.array(wnum, dtype=np.float64)
        return (C2 * wnum) / np.log(1.0 + (C1 * wnum**3 / radiance))

    if temperature is not None:
        temperature = np.array(temperature, dtype=np.float64)
        wnum = np.array(wnum, dtype=np.float64)
        return (C1 * wnum**3) / (np.exp(C2 * wnum / temperature) - 1.0)



"""
=============================
act.discovery (act.discovery)
=============================

.. currentmodule:: act.discovery

This module contains procedures for exploring and downloading data on
ARM Data Discovery.

.. autosummary::
    :toctree: generated/

    download_data
    croptype
    get_asos
"""

from .get_armfiles import download_data
from .get_CropScape import croptype
from .get_asos import get_asos



import pytest
import numpy as np
import glob
import xarray as xr
import pandas as pd
import os
import act
import act.io.armfiles as arm
import act.tests.sample_files as sample_files
from act.plotting import TimeSeriesDisplay, WindRoseDisplay
from act.plotting import SkewTDisplay, XSectionDisplay
from act.plotting import GeographicPlotDisplay, HistogramDisplay
from act.plotting import ContourDisplay
from act.utils.data_utils import accumulate_precip
import matplotlib
matplotlib.use('Agg')
try:
    import metpy.calc as mpcalc
    METPY = True
except ImportError:
    METPY = False


@pytest.mark.mpl_image_compare(tolerance=30)
def test_plot():
    # Process MET data to get simple LCL
    files = sample_files.EXAMPLE_MET_WILDCARD
    met = arm.read_netcdf(files)
    met_temp = met.temp_mean
    met_rh = met.rh_mean
    met_lcl = (20. + met_temp / 5.) * (100. - met_rh) / 1000.
    met['met_lcl'] = met_lcl * 1000.
    met['met_lcl'].attrs['units'] = 'm'
    met['met_lcl'].attrs['long_name'] = 'LCL Calculated from SGP MET E13'

    # Plot data
    display = TimeSeriesDisplay(met)
    display.add_subplots((2, 2), figsize=(15, 10))
    display.plot('wspd_vec_mean', subplot_index=(0, 0))
    display.plot('temp_mean', subplot_index=(1, 0))
    display.plot('rh_mean', subplot_index=(0, 1))

    windrose = WindRoseDisplay(met)
    display.put_display_in_subplot(windrose, subplot_index=(1, 1))
    windrose.plot('wdir_vec_mean', 'wspd_vec_mean',
                  spd_bins=np.linspace(0, 10, 4))
    windrose.axes[0].legend(loc='best')
    met.close()
    return display.fig


def test_errors():
    files = sample_files.EXAMPLE_MET_WILDCARD
    obj = arm.read_netcdf(files)

    display = TimeSeriesDisplay(obj)
    display.axes = None
    with np.testing.assert_raises(RuntimeError):
        display.day_night_background()

    display = TimeSeriesDisplay({'met': obj, 'met2': obj})
    with np.testing.assert_raises(ValueError):
        display.plot('temp_mean')
    with np.testing.assert_raises(ValueError):
        display.qc_flag_block_plot('qc_temp_mean')
    with np.testing.assert_raises(ValueError):
        display.plot_barbs_from_spd_dir('wdir_vec_mean', 'wspd_vec_mean')
    with np.testing.assert_raises(ValueError):
        display.plot_barbs_from_u_v('wdir_vec_mean', 'wspd_vec_mean')

    del obj.attrs['_file_dates']

    data = np.empty(len(obj['time'])) * np.nan
    lat = obj['lat'].values
    lon = obj['lon'].values
    obj['lat'].values = data
    obj['lon'].values = data

    display = TimeSeriesDisplay(obj)
    display.plot('temp_mean')
    display.set_yrng([0, 0])
    with np.testing.assert_warns(RuntimeWarning):
        display.day_night_background()
    obj['lat'].values = lat
    with np.testing.assert_warns(RuntimeWarning):
        display.day_night_background()
    obj['lon'].values = lon * 100.
    with np.testing.assert_warns(RuntimeWarning):
        display.day_night_background()
    obj['lat'].values = lat * 100.
    with np.testing.assert_warns(RuntimeWarning):
        display.day_night_background()

    obj.close()

    # Test some of the other errors
    obj = arm.read_netcdf(files)
    del obj['temp_mean'].attrs['units']
    display = TimeSeriesDisplay(obj)
    display.axes = None
    with np.testing.assert_raises(RuntimeError):
        display.set_yrng([0, 10])
    with np.testing.assert_raises(RuntimeError):
        display.set_xrng([0, 10])
    display.fig = None
    display.plot('temp_mean', add_nan=True)

    assert display.fig is not None
    assert display.axes is not None

    with np.testing.assert_raises(AttributeError):
        display = TimeSeriesDisplay([])

    fig, ax = matplotlib.pyplot.subplots()
    display = TimeSeriesDisplay(obj)
    display.add_subplots((2, 2), figsize=(15, 10))
    display.assign_to_figure_axis(fig, ax)
    assert display.fig is not None
    assert display.axes is not None

    obj = arm.read_netcdf(files)
    display = TimeSeriesDisplay(obj)
    obj.clean.cleanup()
    display.axes = None
    display.fig = None
    display.qc_flag_block_plot('atmos_pressure')
    assert display.fig is not None
    assert display.axes is not None


def test_histogram_errors():
    files = sample_files.EXAMPLE_MET1
    obj = arm.read_netcdf(files)

    histdisplay = HistogramDisplay(obj)
    histdisplay.axes = None
    with np.testing.assert_raises(RuntimeError):
        histdisplay.set_yrng([0, 10])
    with np.testing.assert_raises(RuntimeError):
        histdisplay.set_xrng([-40, 40])
    histdisplay.fig = None
    histdisplay.plot_stacked_bar_graph('temp_mean', bins=np.arange(-40, 40, 5))
    histdisplay.set_yrng([0, 0])
    assert histdisplay.yrng[0][1] == 1.
    assert histdisplay.fig is not None
    assert histdisplay.axes is not None

    with np.testing.assert_raises(AttributeError):
        HistogramDisplay([])

    histdisplay.axes = None
    histdisplay.fig = None
    histdisplay.plot_stairstep_graph('temp_mean', bins=np.arange(-40, 40, 5))
    assert histdisplay.fig is not None
    assert histdisplay.axes is not None

    sigma = 10
    mu = 50
    bins = np.linspace(0, 100, 50)
    ydata = 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(-(bins - mu)**2 / (2 * sigma**2))
    y_array = xr.DataArray(ydata, dims={'bins': bins})
    bins = xr.DataArray(bins, dims={'bins': bins})
    my_fake_ds = xr.Dataset({'bins': bins, 'ydata': y_array})
    histdisplay = HistogramDisplay(my_fake_ds)
    histdisplay.axes = None
    histdisplay.fig = None
    histdisplay.plot_size_distribution('ydata', 'bins', set_title='Fake distribution.')
    assert histdisplay.fig is not None
    assert histdisplay.axes is not None

    sonde_ds = arm.read_netcdf(sample_files.EXAMPLE_SONDE1)
    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.axes = None
    histdisplay.fig = None
    histdisplay.plot_heatmap('tdry', 'alt', x_bins=np.arange(-60, 10, 1),
                             y_bins=np.linspace(0, 10000., 50), cmap='coolwarm')
    assert histdisplay.fig is not None
    assert histdisplay.axes is not None


def test_xsection_errors():
    obj = arm.read_netcdf(sample_files.EXAMPLE_CEIL1)

    display = XSectionDisplay(obj, figsize=(10, 8), subplot_shape=(2,))
    display.axes = None
    with np.testing.assert_raises(RuntimeError):
        display.set_yrng([0, 10])
    with np.testing.assert_raises(RuntimeError):
        display.set_xrng([-40, 40])

    display = XSectionDisplay(obj, figsize=(10, 8), subplot_shape=(1,))
    with np.testing.assert_raises(RuntimeError):
        display.plot_xsection(None, 'backscatter', x='time')

    obj.close()


@pytest.mark.mpl_image_compare(tolerance=30)
def test_multidataset_plot_tuple():
    obj = arm.read_netcdf(sample_files.EXAMPLE_MET1)
    obj2 = arm.read_netcdf(sample_files.EXAMPLE_SIRS)
    obj = obj.rename({'lat': 'fun_time'})
    obj['fun_time'].attrs['standard_name'] = 'latitude'
    obj = obj.rename({'lon': 'not_so_fun_time'})
    obj['not_so_fun_time'].attrs['standard_name'] = 'longitude'

    # You can use tuples if the datasets in the tuple contain a
    # datastream attribute. This is required in all ARM datasets.
    display = TimeSeriesDisplay(
        (obj, obj2), subplot_shape=(2,), figsize=(15, 10))
    display.plot('short_direct_normal', 'sgpsirsE13.b1', subplot_index=(0,))
    display.day_night_background('sgpsirsE13.b1', subplot_index=(0,))
    display.plot('temp_mean', 'sgpmetE13.b1', subplot_index=(1,))
    display.day_night_background('sgpmetE13.b1', subplot_index=(1,))

    ax = act.plotting.common.parse_ax(ax=None)
    ax, fig = act.plotting.common.parse_ax_fig(ax=None, fig=None)
    obj.close()
    obj2.close()
    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_multidataset_plot_dict():

    obj = arm.read_netcdf(sample_files.EXAMPLE_MET1)
    obj2 = arm.read_netcdf(sample_files.EXAMPLE_SIRS)

    # You can use tuples if the datasets in the tuple contain a
    # datastream attribute. This is required in all ARM datasets.
    display = TimeSeriesDisplay(
        {'sirs': obj2, 'met': obj},
        subplot_shape=(2,), figsize=(15, 10))
    display.plot('short_direct_normal', 'sirs', subplot_index=(0,))
    display.day_night_background('sirs', subplot_index=(0,))
    display.plot('temp_mean', 'met', subplot_index=(1,))
    display.day_night_background('met', subplot_index=(1,))
    obj.close()
    obj2.close()

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_wind_rose():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_TWP_SONDE_WILDCARD)

    WindDisplay = WindRoseDisplay(sonde_ds, figsize=(10, 10))
    WindDisplay.plot('deg', 'wspd',
                     spd_bins=np.linspace(0, 20, 10), num_dirs=30,
                     tick_interval=2, cmap='viridis')
    WindDisplay.set_thetarng(trng=(0., 360.))
    WindDisplay.set_rrng((0., 14))

    sonde_ds.close()
    return WindDisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_barb_sounding_plot():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_TWP_SONDE_WILDCARD)
    BarbDisplay = TimeSeriesDisplay({'sonde_darwin': sonde_ds})
    BarbDisplay.plot_time_height_xsection_from_1d_data('rh', 'pres',
                                                       cmap='coolwarm_r',
                                                       vmin=0, vmax=100,
                                                       num_time_periods=25)
    BarbDisplay.plot_barbs_from_spd_dir('deg', 'wspd', 'pres',
                                        num_barbs_x=20)
    sonde_ds.close()
    return BarbDisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_skewt_plot():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    if METPY:
        skewt = SkewTDisplay(sonde_ds)
        skewt.plot_from_u_and_v('u_wind', 'v_wind', 'pres', 'tdry', 'dp')
        sonde_ds.close()
        return skewt.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_skewt_plot_spd_dir():
    sonde_ds = arm.read_netcdf(sample_files.EXAMPLE_SONDE1)

    if METPY:
        skewt = SkewTDisplay(sonde_ds, ds_name='act_datastream')
        skewt.plot_from_spd_and_dir('wspd', 'deg', 'pres', 'tdry', 'dp')
        sonde_ds.close()
        return skewt.fig


@pytest.mark.mpl_image_compare(tolerance=67)
def test_multi_skewt_plot():

    files = glob.glob(sample_files.EXAMPLE_TWP_SONDE_20060121)
    test = {}
    for f in files:
        time = f.split('.')[-3]
        sonde_ds = arm.read_netcdf(f)
        test.update({time: sonde_ds})

    if METPY:
        skewt = SkewTDisplay(test, subplot_shape=(2, 2))
        i = 0
        j = 0
        for f in files:
            time = f.split('.')[-3]
            skewt.plot_from_spd_and_dir('wspd', 'deg', 'pres', 'tdry', 'dp',
                                        subplot_index=(j, i), dsname=time,
                                        p_levels_to_plot=np.arange(10., 1000., 25))
            if j == 1:
                i += 1
                j = 0
            elif j == 0:
                j += 1
        return skewt.fig


@pytest.mark.mpl_image_compare(tolerance=31)
def test_xsection_plot():
    visst_ds = arm.read_netcdf(
        sample_files.EXAMPLE_CEIL1)

    xsection = XSectionDisplay(visst_ds, figsize=(10, 8))
    xsection.plot_xsection(None, 'backscatter', x='time', y='range',
                           cmap='coolwarm', vmin=0, vmax=320)
    visst_ds.close()
    return xsection.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_xsection_plot_map():
    radar_ds = arm.read_netcdf(
        sample_files.EXAMPLE_VISST, combine='nested')
    try:
        xsection = XSectionDisplay(radar_ds, figsize=(15, 8))
        xsection.plot_xsection_map(None, 'ir_temperature', vmin=220, vmax=300, cmap='Greys',
                                   x='longitude', y='latitude', isel_kwargs={'time': 0})
        radar_ds.close()
        return xsection.fig
    except Exception:
        pass


@pytest.mark.mpl_image_compare(tolerance=30)
def test_geoplot():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)
    try:
        geodisplay = GeographicPlotDisplay({'sgpsondewnpnC1.b1': sonde_ds})
        geodisplay.geoplot('tdry', marker='.', cartopy_feature=['STATES', 'LAND', 'OCEAN', 'COASTLINE',
                                                                'BORDERS', 'LAKES', 'RIVERS'],
                           text={'Ponca City': [-97.0725, 36.7125]})
        return geodisplay.fig
    except Exception:
        pass
    sonde_ds.close()


@pytest.mark.mpl_image_compare(tolerance=30)
def test_stair_graph():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_stairstep_graph('tdry', bins=np.arange(-60, 10, 1))
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_stair_graph_sorted():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_stairstep_graph(
        'tdry', bins=np.arange(-60, 10, 1), sortby_field="alt",
        sortby_bins=np.linspace(0, 10000., 6))
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_stacked_bar_graph():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_stacked_bar_graph('tdry', bins=np.arange(-60, 10, 1))
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_stacked_bar_graph2():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_stacked_bar_graph('tdry')
    histdisplay.set_yrng([0, 400])
    histdisplay.set_xrng([-70, 0])
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_stacked_bar_graph_sorted():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_stacked_bar_graph(
        'tdry', bins=np.arange(-60, 10, 1), sortby_field="alt",
        sortby_bins=np.linspace(0, 10000., 6))
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_heatmap():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    histdisplay = HistogramDisplay({'sgpsondewnpnC1.b1': sonde_ds})
    histdisplay.plot_heatmap(
        'tdry', 'alt', x_bins=np.arange(-60, 10, 1),
        y_bins=np.linspace(0, 10000., 50), cmap='coolwarm')
    sonde_ds.close()

    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_size_distribution():
    sigma = 10
    mu = 50
    bins = np.linspace(0, 100, 50)
    ydata = 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(-(bins - mu)**2 / (2 * sigma**2))
    y_array = xr.DataArray(ydata, dims={'bins': bins})
    bins = xr.DataArray(bins, dims={'bins': bins})
    my_fake_ds = xr.Dataset({'bins': bins, 'ydata': y_array})
    histdisplay = HistogramDisplay(my_fake_ds)
    histdisplay.plot_size_distribution('ydata', 'bins', set_title='Fake distribution.')
    return histdisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_contour():
    files = glob.glob(sample_files.EXAMPLE_MET_CONTOUR)
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        obj = arm.read_netcdf(f)
        data.update({f: obj})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50,
                           contour='contour', cmap='viridis')
    display.plot_vectors_from_spd_dir(fields=wind_fields, time=time, mesh=True, grid_delta=(0.1, 0.1))
    display.plot_station(fields=station_fields, time=time, markersize=7, color='red')

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_contour_stamp():
    files = glob.glob(sample_files.EXAMPLE_STAMP_WILDCARD)
    test = {}
    stamp_fields = {}
    time = '2020-01-01T00:00:00.000000000'
    for f in files:
        ds = f.split('/')[-1]
        obj = act.io.armfiles.read_netcdf(f)
        test.update({ds: obj})
        stamp_fields.update({ds: ['lon', 'lat', 'plant_water_availability_east']})
        obj.close()

    display = act.plotting.ContourDisplay(test, figsize=(8, 8))
    display.create_contour(fields=stamp_fields, time=time, levels=50,
                           alpha=0.5, twod_dim_value=5)

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_contour2():
    files = glob.glob(sample_files.EXAMPLE_MET_CONTOUR)
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        obj = arm.read_netcdf(f)
        data.update({f: obj})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50,
                           contour='contour', cmap='viridis')
    display.plot_vectors_from_spd_dir(fields=wind_fields, time=time, mesh=False, grid_delta=(0.1, 0.1))
    display.plot_station(fields=station_fields, time=time, markersize=7, color='pink')

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_contourf():
    files = glob.glob(sample_files.EXAMPLE_MET_CONTOUR)
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        obj = arm.read_netcdf(f)
        data.update({f: obj})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure', 'temp_mean', 'rh_mean',
                                   'vapor_pressure_mean', 'temp_std']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50,
                           contour='contourf', cmap='viridis')
    display.plot_vectors_from_spd_dir(fields=wind_fields, time=time, mesh=True, grid_delta=(0.1, 0.1))
    display.plot_station(fields=station_fields, time=time, markersize=7, color='red')

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_contourf2():
    files = glob.glob(sample_files.EXAMPLE_MET_CONTOUR)
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        obj = arm.read_netcdf(f)
        data.update({f: obj})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure', 'temp_mean', 'rh_mean',
                                   'vapor_pressure_mean', 'temp_std']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50,
                           contour='contourf', cmap='viridis')
    display.plot_vectors_from_spd_dir(fields=wind_fields, time=time, mesh=False, grid_delta=(0.1, 0.1))
    display.plot_station(fields=station_fields, time=time, markersize=7, color='pink')

    return display.fig


# Due to issues with pytest-mpl, for now we just test to see if it runs
def test_time_height_scatter():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_SONDE1)

    display = TimeSeriesDisplay({'sgpsondewnpnC1.b1': sonde_ds},
                                figsize=(7, 3))
    display.time_height_scatter('tdry', day_night_background=False)

    sonde_ds.close()

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_qc_bar_plot():
    ds_object = arm.read_netcdf(sample_files.EXAMPLE_MET1)
    ds_object.clean.cleanup()
    var_name = 'temp_mean'
    ds_object.qcfilter.set_test(var_name, index=range(100, 600), test_number=2)

    # Testing out when the assessment is not listed
    ds_object.qcfilter.set_test(var_name, index=range(500, 800), test_number=4)
    ds_object['qc_' + var_name].attrs['flag_assessments'][3] = 'Wonky'

    display = TimeSeriesDisplay({'sgpmetE13.b1': ds_object},
                                subplot_shape=(2, ), figsize=(7, 4))
    display.plot(var_name, subplot_index=(0, ), assessment_overplot=True)
    display.day_night_background('sgpmetE13.b1', subplot_index=(0, ))
    color_lookup = {'Bad': 'red', 'Incorrect': 'red',
                    'Indeterminate': 'orange', 'Suspect': 'orange',
                    'Missing': 'darkgray', 'Not Failing': 'green',
                    'Acceptable': 'green'}
    display.qc_flag_block_plot(var_name, subplot_index=(1, ),
                               assessment_color=color_lookup)

    ds_object.close()

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_2d_as_1d():
    obj = arm.read_netcdf(sample_files.EXAMPLE_CEIL1)

    display = TimeSeriesDisplay(obj)
    display.plot('backscatter', force_line_plot=True)

    obj.close()
    del obj

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_fill_between():
    obj = arm.read_netcdf(sample_files.EXAMPLE_MET_WILDCARD)

    accumulate_precip(obj, 'tbrg_precip_total')

    display = TimeSeriesDisplay(obj)
    display.fill_between('tbrg_precip_total_accumulated', color='gray', alpha=0.2)

    obj.close()
    del obj

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_qc_flag_block_plot():
    obj = arm.read_netcdf(sample_files.EXAMPLE_SURFSPECALB1MLAWER)

    display = TimeSeriesDisplay(obj, subplot_shape=(2, ), figsize=(8, 2 * 4))

    display.plot('surface_albedo_mfr_narrowband_10m', force_line_plot=True, labels=True)

    display.qc_flag_block_plot('surface_albedo_mfr_narrowband_10m', subplot_index=(1, ))

    obj.close()
    del obj

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_assessment_overplot():
    var_name = 'temp_mean'
    files = sample_files.EXAMPLE_MET1
    ds = arm.read_netcdf(files)
    ds.load()
    ds.clean.cleanup()

    ds.qcfilter.set_test(var_name, index=np.arange(100, 300, dtype=int), test_number=2)
    ds.qcfilter.set_test(var_name, index=np.arange(420, 422, dtype=int), test_number=3)
    ds.qcfilter.set_test(var_name, index=np.arange(500, 800, dtype=int), test_number=4)
    ds.qcfilter.set_test(var_name, index=np.arange(900, 901, dtype=int), test_number=4)

    # Plot data
    display = TimeSeriesDisplay(ds, subplot_shape=(1, ), figsize=(10, 6))
    display.plot(var_name, day_night_background=True, assessment_overplot=True)

    ds.close()
    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_assessment_overplot_multi():
    var_name1, var_name2 = 'wspd_arith_mean', 'wspd_vec_mean'
    files = sample_files.EXAMPLE_MET1
    ds = arm.read_netcdf(files)
    ds.load()
    ds.clean.cleanup()

    ds.qcfilter.set_test(var_name1, index=np.arange(100, 200, dtype=int), test_number=2)
    ds.qcfilter.set_test(var_name1, index=np.arange(500, 600, dtype=int), test_number=4)
    ds.qcfilter.set_test(var_name2, index=np.arange(300, 400, dtype=int), test_number=4)

    # Plot data
    display = TimeSeriesDisplay(ds, subplot_shape=(1, ), figsize=(10, 6))
    display.plot(var_name1, label=var_name1,
                 assessment_overplot=True, overplot_behind=True)
    display.plot(var_name2, day_night_background=True, color='green',
                 label=var_name2, assessment_overplot=True)

    ds.close()
    return display.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_plot_barbs_from_u_v():
    sonde_ds = arm.read_netcdf(
        sample_files.EXAMPLE_TWP_SONDE_WILDCARD)
    BarbDisplay = TimeSeriesDisplay({'sonde_darwin': sonde_ds})
    BarbDisplay.plot_barbs_from_u_v('u_wind', 'v_wind', 'pres',
                                    num_barbs_x=20)
    sonde_ds.close()
    return BarbDisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_plot_barbs_from_u_v2():
    bins = list(np.linspace(0, 1, 10))
    xbins = list(pd.date_range(pd.to_datetime('2020-01-01'),
                               pd.to_datetime('2020-01-02'), 12))
    y_data = np.full([len(xbins), len(bins)], 1.)
    x_data = np.full([len(xbins), len(bins)], 2.)
    y_array = xr.DataArray(y_data, dims={'xbins': xbins, 'ybins': bins},
                           attrs={'units': 'm/s'})
    x_array = xr.DataArray(x_data, dims={'xbins': xbins, 'ybins': bins},
                           attrs={'units': 'm/s'})
    xbins = xr.DataArray(xbins, dims={'xbins': xbins})
    ybins = xr.DataArray(bins, dims={'ybins': bins})
    fake_obj = xr.Dataset({'xbins': xbins, 'ybins': ybins,
                          'ydata': y_array, 'xdata': x_array})
    BarbDisplay = TimeSeriesDisplay(fake_obj)
    BarbDisplay.plot_barbs_from_u_v('xdata', 'ydata', None, num_barbs_x=20,
                                    num_barbs_y=20, set_title='test plot', cmap='jet')
    fake_obj.close()
    return BarbDisplay.fig


@pytest.mark.mpl_image_compare(tolerance=30)
def test_2D_timeseries_plot():
    obj = arm.read_netcdf(sample_files.EXAMPLE_CEIL1)
    display = TimeSeriesDisplay(obj)
    display.plot('backscatter', y_rng=[0, 5000], use_var_for_y='range')
    matplotlib.pyplot.show()
    return display.fig



"""
act.qc.clean
------------------------------

Class definitions for cleaning up QC variables to standard
cf-compliance

"""

import xarray as xr
import re
import numpy as np
import copy


@xr.register_dataset_accessor('clean')
class CleanDataset(object):
    """
    Class for cleaning up QC variables to standard cf-compliance
    """
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    @property
    def matched_qc_variables(self, check_arm_syntax=True):
        """
        Find variables that are QC variables and return list of names.

        Parameters
        ----------
        check_arm_syntax : boolean
            ARM ueses a standard of starting all quality control variables
            with "qc" joined with an underscore. This is a more robust method
            of getting the quality control variables before the standard_name
            attribute is added. If this is true will first check using
            attributes and will then check if variable starts with "qc".

        Returns
        -------
        variables : list of str
            A list of strings containing the name of each variable.

        """
        variables = []

        # Will need to find all historical cases and add to list
        qc_dict = {'description':
                   ["See global attributes for individual.+bit descriptions.",
                    ("This field contains bit packed integer values, where each "
                     "bit represents a QC test on the data. Non-zero bits indicate "
                     "the QC condition given in the description for those bits; "
                     "a value of 0.+ indicates the data has not "
                     "failed any QC tests."),
                    (r"This field contains bit packed values which should be "
                     r"interpreted as listed..+")
                    ]
                   }

        # Loop over each variable and look for a match to an attribute that
        # would exist if the variable is a QC variable
        for var in self._obj.data_vars:
            attributes = self._obj[var].attrs
            for att_name in attributes:
                if att_name in qc_dict.keys():
                    for value in qc_dict[att_name]:
                        if re.match(value, attributes[att_name]) is not None:
                            variables.append(var)
                            break

        # Check the start of the variable name. If it begins with qc_ assume quality
        # control variable from ARM.
        if check_arm_syntax:
            variables_qc = [var for var in self._obj.data_vars if var.startswith('qc_')]
            variables = variables + variables_qc
            variables = list(set(variables))

        return variables

    def cleanup(self, cleanup_arm_qc=True, clean_arm_state_vars=None,
                handle_missing_value=True, link_qc_variables=True,
                normalize_assessment=False,
                **kwargs):
        """
        Wrapper method to automatically call all the standard methods
        for obj cleanup.

        Parameters
        ----------
        cleanup_arm_qc : bool
            Option to clean xarray object from ARM QC to CF QC standards.
            Default is True.
        clean_arm_state_vars : list of str
            Option to clean xarray object state variables from ARM to CF
            standards. Pass in list of variable names.
        handle_missing_value : bool
            Go through variables and look for cases where a QC or state varible
            was convereted to a float and missing values set to np.nan. This
            is done because of xarry's default to use mask_and_scale=True.
            This will convert the data type back to integer and replace
            any instances of np.nan to a missing value indicator (most
            likely -9999).
        link_qc_variables : bool
            Option to link QC variablers through ancillary_variables if not
            already set.
        normalize_assessment : bool
            Option to clean up assessments to use the same terminology. Set to
            False for default because should only be an issue after adding DQRs
            and the function to add DQRs calls this method.
        **kwargs : keywords
            Keyword arguments passed through to clean.clean_arm_qc
            method.

        """
        # Convert ARM QC to be more like CF state fields
        if cleanup_arm_qc:
            self._obj.clean.clean_arm_qc(**kwargs)

        # Convert ARM state fields to be more liek CF state fields
        if clean_arm_state_vars is not None:
            self._obj.clean.clean_arm_state_variables(clean_arm_state_vars)

        # Correctly convert data type because of missing value
        # indicators in state and QC variables. Needs to be run after
        # clean.clean_arm_qc to use CF attribute names.
        if handle_missing_value:
            self._obj.clean.handle_missing_values()

        # Add some ancillary_variables linkages
        # between data variable and QC variable
        if link_qc_variables:
            self._obj.clean.link_variables()

        # Update the terminology used with flag_assessments to be consistent
        if normalize_assessment:
            self._obj.clean.normalize_assessment()

    def handle_missing_values(self, default_missing_value=np.int32(-9999)):
        """
        Correctly handle missing_value and _FillValue in object.
        xarray will automatically replace missing_value and
        _FillValue in the data with NaN. This is great for data set
        as type float but not great for int data. Can cause issues
        with QC and state fields. This will loop through the array
        looking for state and QC fields and revert them back to int
        data type if upconverted to float to handle NaNs. Issue is that
        xarray will convert data type to float if the attribute is defined
        even if no data are set as missing value. xarray will also then
        remove the missing_value or _FillValue variable attribute. This
        will put the missing_value attribute back if needed.

        Parameters
        ----------
        default_missing_value : numpy int or float
           The default missing value to use if a missing_value attribute
           is not defined but one is needed.

        """
        state_att_names = ['flag_values', 'flag_meanings',
                           'flag_masks', 'flag_attributes']

        # Look for variables that have 2 of the state_att_names defined
        # as attribures and is of type float. If so assume the variable
        # was incorreclty converted to float type.
        for var in self._obj.data_vars:
            var_att_names = self._obj[var].attrs.keys()
            if (len(set(state_att_names) & set(var_att_names)) >= 2 and
                self._obj[var].values.dtype in
                [np.dtype('float16'), np.dtype('float32'),
                 np.dtype('float64')]):

                # Look at units variable to see if this is the stupid way some
                # ARM products mix data and state variables. If the units are not
                # in the normal list of unitless type assume this is a data variable
                # and skip. Other option is to lookf or a valid_range attribute
                # and skip. This is commented out for now since the units check
                # appears to be working.
                try:
                    if self._obj[var].attrs['units'] not in ['1', 'unitless', '', ' ']:
                        continue
#                    self._obj[var].attrs['valid_range']
#                    continue
                except KeyError:
                    pass

                # Change any np.nan values to missing value indicator
                data = self._obj[var].values
                data[np.isnan(data)] = default_missing_value.astype(data.dtype)

                # Convert data to match type of flag_mask or flag_values
                # as the best guess of what type is correct.
                found_dtype = False
                for att_name in ['flag_masks', 'flag_values']:
                    try:
                        att_value = self._obj[var].attrs[att_name]
                        if isinstance(att_value, (list, tuple)):
                            dtype = att_value[0].dtype
                        else:
                            dtype = att_value.dtype
                        data = data.astype(dtype)
                        found_dtype = True
                        break
                    except (KeyError, IndexError):
                        pass

                # If flag_mask or flag_values is not available choose an int type
                # and set data to that type.
                if found_dtype is False:
                    data = data.astype(default_missing_value.dtype)

                # Return data to object and add missing value indicator
                # attribute to variable.
                self._obj[var].values = data
                self._obj[var].attrs['missing_value'] = \
                    default_missing_value.astype(data.dtype)

    def get_attr_info(self, variable=None, flag=False):
        """
        Get ARM quality control definitions from the ARM standard
        bit_#_description, ... attributes and return as dictionary.
        Will attempt to guess if the flag is integer or bit packed
        based on what attributes are set.

        Parameters
        ----------
        variable : str
            Variable name to get attribute information. If set to None
            will get global attributes.
        flag : bool
            Optional flag indicating if QC is expected to be bitpacked
            or integer. Flag = True indicates integer QC. Default
            is bitpacked or False.

        Returns
        -------
        attributes dictionary : dict or None
            A dictionary contianing the attribute information converted from
            ARM QC to CF QC. All keys include 'flag_meanings', 'flag_masks',
            'flag_values', 'flag_assessments', 'flag_tests', 'arm_attributes'.
            Returns None if none found.

        """
        string = 'bit'
        if flag:
            string = 'flag'
        else:
            found_string = False
            try:
                if self._obj.attrs['qc_bit_comment']:
                    string = 'bit'
                    found_string = True
            except KeyError:
                pass

            if found_string is False:
                try:
                    if self._obj.attrs['qc_flag_comment']:
                        string = 'flag'
                        found_string = True
                except KeyError:
                    pass

            if found_string is False:
                var = self.matched_qc_variables
                if len(var) > 0:
                    try:
                        if self._obj[variable].attrs['flag_method'] == 'integer':
                            string = 'flag'
                        found_string = True
                        del self._obj[variable].attrs['flag_method']
                    except KeyError:
                        pass

        try:
            if variable:
                attr_description_pattern = (r"(^" + string +
                                            r")_([0-9]+)_(description$)")
                attr_assessment_pattern = (r"(^" + string +
                                           r")_([0-9]+)_(assessment$)")
                attr_comment_pattern = (r"(^" + string +
                                        r")_([0-9]+)_(comment$)")
                attributes = self._obj[variable].attrs
            else:
                attr_description_pattern = (r"(^qc_" + string +
                                            r")_([0-9]+)_(description$)")
                attr_assessment_pattern = (r"(^qc_" + string +
                                           r")_([0-9]+)_(assessment$)")
                attr_comment_pattern = (r"(^qc_" + string +
                                        r")_([0-9]+)_(comment$)")
                attributes = self._obj.attrs
        except KeyError:
            return None

        assessment_bit_num = []
        description_bit_num = []
        comment_bit_num = []
        flag_masks = []
        flag_meanings = []
        flag_assessments = []
        flag_comments = []
        arm_attributes = []

        dtype = np.int32
        for att_name in attributes:
            try:
                description = re.match(attr_description_pattern, att_name)
                description_bit_num.append(int(description.groups()[1]))
                flag_meanings.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

            try:
                assessment = re.match(attr_assessment_pattern, att_name)
                assessment_bit_num.append(int(assessment.groups()[1]))
                flag_assessments.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

            try:
                comment = re.match(attr_comment_pattern, att_name)
                comment_bit_num.append(int(comment.groups()[1]))
                flag_comments.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

        if variable is not None:
            # Try and get the data type from the variable if it is an int
            try:
                if (self._obj[variable].values.dtype in [
                        np.dtype('int8'), np.dtype('int16'),
                        np.dtype('int32'), np.dtype('int64')]):
                    dtype = self._obj[variable].values.dtype
            except AttributeError:
                pass

            # If the data is type float check the largest value and make
            # sure the type we set can handle it.
            if np.nanmax(self._obj[variable].values) > 2**32 - 1:
                dtype = np.int64

        # Sort on bit number to ensure correct description order
        index = np.argsort(description_bit_num)
        flag_meanings = np.array(flag_meanings)
        description_bit_num = np.array(description_bit_num)
        flag_meanings = flag_meanings[index]
        description_bit_num = description_bit_num[index]

        # Sort on bit number to ensure correct assessment order
        if len(flag_assessments) > 0:
            if len(flag_assessments) < len(flag_meanings):
                for ii in range(1, len(flag_meanings) + 1):
                    if ii not in assessment_bit_num:
                        assessment_bit_num.append(ii)
                        flag_assessments.append('')
            index = np.argsort(assessment_bit_num)
            flag_assessments = np.array(flag_assessments)
            flag_assessments = flag_assessments[index]

        # Sort on bit number to ensure correct comment order
        if len(flag_comments) > 0:
            if len(flag_comments) < len(flag_meanings):
                for ii in range(1, len(flag_meanings) + 1):
                    if ii not in comment_bit_num:
                        comment_bit_num.append(ii)
                        flag_comments.append('')
            index = np.argsort(comment_bit_num)
            flag_comments = np.array(flag_comments)
            flag_comments = flag_comments[index]

        # Convert bit number to mask number
        if len(description_bit_num) > 0:
            flag_masks = np.array(description_bit_num)
            flag_masks = np.left_shift(1, flag_masks - 1)

        # build dictionary to return values
        if len(flag_masks) > 0 or len(description_bit_num) > 0:
            return_dict = dict()
            return_dict['flag_meanings'] = list(np.array(flag_meanings,
                                                         dtype=str))
            if len(flag_masks) > 0 and max(flag_masks) > 2**32 - 1:
                flag_mask_dtype = np.int64
            else:
                flag_mask_dtype = dtype

            if flag:
                return_dict['flag_values'] = list(np.array(description_bit_num,
                                                           dtype=dtype))
                return_dict['flag_masks'] = list(np.array([],
                                                          dtype=flag_mask_dtype))
            else:
                return_dict['flag_values'] = list(np.array([],
                                                           dtype=dtype))
                return_dict['flag_masks'] = list(np.array(flag_masks,
                                                          dtype=flag_mask_dtype))

            return_dict['flag_assessments'] = list(np.array(flag_assessments,
                                                            dtype=str))
            return_dict['flag_tests'] = list(np.array(description_bit_num,
                                                      dtype=dtype))
            return_dict['flag_comments'] = list(np.array(flag_comments,
                                                         dtype=str))
            return_dict['arm_attributes'] = arm_attributes

        else:
            # If nothing to return set to None
            return_dict = None

        return return_dict

    def clean_arm_state_variables(self,
                                  variables,
                                  override_cf_flag=True,
                                  clean_units_string=True,
                                  integer_flag=True):
        """
        Function to clean up state variables to use more CF style.

        Parameters
        ----------
        variables : str or list of str
            List of variable names to update.
        override_cf_flag : bool
            Option to overwrite CF flag_meanings attribute if it exists
            with the values from ARM QC bit_#_description.
        clean_units_string : bool
            Option to update units string if set to 'unitless' to be
            udunits compliant '1'.
        integer_flag : bool
            Pass through keyword of 'flag' for get_attr_info().

        """
        if isinstance(variables, str):
            variables = [variables]

        for var in variables:
            flag_info = self.get_attr_info(variable=var, flag=integer_flag)
            if flag_info is None:
                return

            # Add new attributes to variable
            for attr in ['flag_values', 'flag_meanings', 'flag_masks']:

                if len(flag_info[attr]) > 0:
                    # Only add if attribute does not exist.
                    if attr in self._obj[var].attrs.keys() is False:
                        self._obj[var].attrs[attr] = copy.copy(flag_info[attr])
                    # If flag is set set attribure even if exists
                    elif override_cf_flag:
                        self._obj[var].attrs[attr] = copy.copy(flag_info[attr])

            # Remove replaced attributes
            arm_attributes = flag_info['arm_attributes']
            for attr in arm_attributes:
                try:
                    del self._obj[var].attrs[attr]
                except KeyError:
                    pass

            # Clean up units attribute from unitless to udunits '1'
            if (clean_units_string and
                    self._obj[var].attrs['units'] == 'unitless'):
                self._obj[var].attrs['units'] = '1'

    def correct_valid_minmax(self, qc_variable):
        """
        Function to correct the name and location of quality control limit
        variables that use valid_min and valid_max incorrectly.

        Parameters
        ----------
        qc_variable : str
            Name of quality control variable in xarray object to correct.

        """
        test_dict = {'valid_min': 'fail_min',
                     'valid_max': 'fail_max',
                     'valid_delta': 'fail_delta'}

        aa = re.match(r"^qc_(.+)", qc_variable)
        variable = None
        try:
            variable = aa.groups()[0]
        except AttributeError:
            return

        made_change = False
        try:
            flag_meanings = copy.copy(
                self._obj[qc_variable].attrs['flag_meanings'])
        except KeyError:
            return

        for attr in test_dict.keys():
            for ii, test in enumerate(flag_meanings):
                if attr in test:
                    flag_meanings[ii] = re.sub(attr, test_dict[attr], test)
                    made_change = True
                    try:
                        self._obj[qc_variable].attrs[test_dict[attr]] = \
                            copy.copy(self._obj[variable].attrs[attr])
                        del self._obj[variable].attrs[attr]
                    except KeyError:
                        pass

        if made_change:
            self._obj[qc_variable].attrs['flag_meanings'] = flag_meanings

    def link_variables(self):
        """
        Add some attributes to link and explain data
        to QC data relationship. Will use non-CF standard_name
        of quality_flag. Hopefully this will be added to the
        standard_name table in the future.
        """
        for var in self._obj.data_vars:
            aa = re.match(r"^qc_(.+)", var)
            try:
                variable = aa.groups()[0]
                qc_variable = var
            except AttributeError:
                continue
            # Skip data quality fields.
            try:
                if not ('Quality check results on field:' in
                        self._obj[var].attrs['long_name']):
                    continue
            except KeyError:
                pass

            # Get existing data variable ancillary_variables attribute
            try:
                ancillary_variables = self._obj[variable].\
                    attrs['ancillary_variables']
            except KeyError:
                ancillary_variables = ''

            # If the QC variable is not in ancillary_variables add
            if qc_variable not in ancillary_variables:
                ancillary_variables = qc_variable
            self._obj[variable].attrs['ancillary_variables']\
                = copy.copy(ancillary_variables)

            # Check if QC variable has correct standard_name and iff not fix it.
            correct_standard_name = 'quality_flag'
            try:
                if self._obj[qc_variable].attrs['standard_name'] != correct_standard_name:
                    self._obj[qc_variable].attrs['standard_name'] = correct_standard_name
            except KeyError:
                self._obj[qc_variable].attrs['standard_name'] = correct_standard_name

    def clean_arm_qc(self,
                     override_cf_flag=True,
                     clean_units_string=True,
                     correct_valid_min_max=True):
        """
        Function to clean up xarray object QC variables.

        Parameters
        ----------
        override_cf_flag : bool
            Option to overwrite CF flag_masks, flag_meanings, flag_values
            if exists.
        clean_units_string : bool
            Option to clean up units string from 'unitless'
            to udunits compliant '1'.
        correct_valid_min_max : bool
            Option to correct use of valid_min and valid_max with QC variables
            by moving from data variable to QC varible, renaming to fail_min,
            fail_max and fail_detla if the valid_min, valid_max or valid_delta
            is listed in bit discription attribute. If not listed as
            used with QC will assume is being used correctly.

        """
        global_qc = self.get_attr_info()
        for qc_var in self.matched_qc_variables:

            # Clean up units attribute from unitless to udunits '1'
            try:
                if (clean_units_string and
                        self._obj[qc_var].attrs['units'] == 'unitless'):
                    self._obj[qc_var].attrs['units'] = '1'
            except KeyError:
                pass

            qc_attributes = self.get_attr_info(variable=qc_var)

            if qc_attributes is None:
                qc_attributes = global_qc

            # Add new attributes to variable
            for attr in ['flag_masks', 'flag_meanings',
                         'flag_assessments', 'flag_values', 'flag_comments']:

                if qc_attributes is not None and len(qc_attributes[attr]) > 0:
                    # Only add if attribute does not exists
                    if attr in self._obj[qc_var].attrs.keys() is False:
                        self._obj[qc_var].attrs[attr] = copy.copy(qc_attributes[attr])
                    # If flag is set add attribure even if already exists
                    elif override_cf_flag:
                        self._obj[qc_var].attrs[attr] = copy.copy(qc_attributes[attr])

            # Remove replaced attributes
            if qc_attributes is not None:
                arm_attributes = qc_attributes['arm_attributes']
                if 'description' not in arm_attributes:
                    arm_attributes.append('description')
                if 'flag_method' not in arm_attributes:
                    arm_attributes.append('flag_method')
                for attr in arm_attributes:
                    try:
                        del self._obj[qc_var].attrs[attr]
                    except KeyError:
                        pass

            # Check for use of valid_min and valid_max as QC limits and fix
            if correct_valid_min_max:
                self._obj.clean.correct_valid_minmax(qc_var)

        # Clean up global attributes
        if global_qc is not None:
            global_attributes = global_qc['arm_attributes']
            global_attributes.extend(['qc_bit_comment'])
            for attr in global_attributes:
                try:
                    del self._obj.attrs[attr]
                except KeyError:
                    pass

    def normalize_assessment(self, variables=None, exclude_variables=None,
                             qc_lookup={"Incorrect": "Bad", "Suspect": "Indeterminate"}):

        """
        Method to clean up assessment terms used to be consistent between
        embedded QC and DQRs.

        Parameters
        ----------
        variables : str or list of str
            Optional data variable names to check and normalize. If set to
            None will check all variables.
        exclude_variables : str or list of str
            Optional data variable names to exclude from processing.
        qc_lookup : dict
            Optional dictionary used to convert between terms.

        """

        # Get list of variables if not provided
        if variables is None:
            variables = list(self._obj.data_vars)

        # Ensure variables is a list
        if not isinstance(variables, (list, tuple)):
            variables = [variables]

        # If exclude variables provided remove from variables list
        if exclude_variables is not None:
            if not isinstance(exclude_variables, (list, tuple)):
                exclude_variables = [exclude_variables]

            variables = list(set(variables) - set(exclude_variables))

        # Loop over variables checking if a QC variable exits and use the
        # lookup dictionary to convert the assessment terms.
        for var_name in variables:
            qc_var_name = self._obj.qcfilter.check_for_ancillary_qc(
                var_name, add_if_missing=False, cleanup=False)
            if qc_var_name is not None:
                try:
                    flag_assessments = self._obj[qc_var_name].attrs['flag_assessments']
                except KeyError:
                    continue

                for ii, assess in enumerate(flag_assessments):
                    try:
                        flag_assessments[ii] = qc_lookup[assess]
                    except KeyError:
                        continue



"""
act.utils.datetime_utils
------------------------

Module that containing utilities involving datetimes.

"""

import datetime as dt
import pandas as pd
import numpy as np
from scipy import stats
import warnings


def dates_between(sdate, edate):
    """
    Ths procedure returns all of the dates between *sdate* and *edate*.

    Parameters
    ----------
    sdate : str
        The string containing the start date. The string is formatted
        YYYYMMDD.
    edate : str
        The string containing the end date. The string is formatted
        YYYYMMDD.

    Returns
    -------
    all_dates : array of datetimes
        The array containing the dates between *sdate* and *edate*.

    """
    days = dt.datetime.strptime(edate, '%Y%m%d') - \
        dt.datetime.strptime(sdate, '%Y%m%d')
    all_dates = [dt.datetime.strptime(sdate, '%Y%m%d') + dt.timedelta(days=d)
                 for d in range(days.days + 1)]
    return all_dates


def numpy_to_arm_date(_date, returnTime=False):
    """
    Given a numpy datetime64, return an ARM standard date (yyyymmdd).

    Parameters
    ----------
    date : numpy.datetime64
        Numpy datetime64 date.
    returnTime : boolean
        If set to true, returns time instead of date

    Returns
    -------
    arm_date : string
        Returns an arm date.

    """
    date = pd.to_datetime(str(_date))
    if returnTime is False:
        date = date.strftime('%Y%m%d')
    else:
        date = date.strftime('%H%M%S')

    return date


def reduce_time_ranges(time, time_delta=60, broken_barh=False):
    """
    Given a time series, this function will return a list of tuples of time
    ranges representing the contineous times where no data is detected missing.

    Parameters
    ----------
    time : numpy datetime64 array
        The numpy array of date time values.
    time_delta : int
        The number of seconds to use as default time step in time array.
    broken_barh : boolean
        Option to return start time and duration instead of start time and
        end time. This is used with the pyplot.broken_barh() plotting routine.

    Returns
    -------
    time_ranges : list of tuples with 2 numpy datetime64 times
        The time range(s) of contineous data.

    """
    # Convert integer sections to numpy datetime64
    time_delta = np.timedelta64(int(time_delta * 1000), 'ms')

    # Make a difference array to find where time difference is great than time_delta
    diff = np.diff(time)
    dd = np.where(diff > time_delta)[0]

    if len(dd) == 0:
        return [(time[0], time[-1] - time[0])]

    # A add to start and end of array for beginning and end values
    dd = np.insert(dd, 0, -1)
    dd = np.append(dd, len(time) - 1)

    # Create a list of tuples containg time ranges or start time with duration
    if broken_barh:
        return [(time[dd[ii] + 1], time[dd[ii + 1]] - time[dd[ii] + 1])
                for ii in range(len(dd) - 1)]
    else:
        return [(time[dd[ii] + 1], time[dd[ii + 1]]) for ii in range(len(dd) - 1)]


def determine_time_delta(time, default=60):
    """
    Returns the most likely time step in seconds by analyzing the difference in
    time steps.

    Parameters
    ----------
    time : numpy datetime64 array
        The numpy array of date time values.
    default : int or float
        The default number to return if unable to calculate a value.

    Returns
    -------
    time_delta : float
        Returns the number of seconds for the most common time step. If can't
        calculate a value the default value is returned.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        if time.size > 1:
            mode = stats.mode(np.diff(time))
            time_delta = mode.mode[0]
            time_delta = time_delta.astype('timedelta64[s]').astype(float)
        else:
            time_delta = default

    return float(time_delta)


def datetime64_to_datetime(time):
    """
    Given a numpy datetime64 array time series, return datetime (y, m, d, h, m, s)

    Parameters
    ----------
    time: numpy datetime64 array, list of numpy datetime64 values or scalar numpy datetime64
        The numpy array of date time values.

    Returns
    -------
    datetime: list
        Returns a list of datetimes (y, m, d, h, m, s) from a time series.

    """
    if isinstance(time, (tuple, list)):
        time = np.array(time)
    if len(time.shape) == 0:
        time = np.array([time])
    datetime_array = [dt.datetime.fromtimestamp(tm.astype('datetime64[ms]').astype('float') / 1000.,
                                                tz=dt.timezone.utc).replace(tzinfo=None) for tm in time]
    return datetime_array
